"""Report on who ingested AIPs and when. An indication is also provided
about how long they took to process.
"""

import pandas as pd
import plotly.express as px
from flask import render_template
from flask import request

from AIPscan.Data import fields
from AIPscan.Data import report_data
from AIPscan.helpers import _simplify_datetime
from AIPscan.helpers import parse_bool
from AIPscan.helpers import parse_datetime_bound
from AIPscan.Reporter import download_csv
from AIPscan.Reporter import get_display_end_date
from AIPscan.Reporter import reporter
from AIPscan.Reporter import request_params
from AIPscan.Reporter import translate_headers

# Response fields.
TRANSFER_COUNT = "transfer_count"
FIGURE_HTML = "figure"

# Error response for the Gantt chart. Because the chart itself is
# returned as HTML we have the opportunity to inject a useful string
# for those using the report incorrectly, e.g. not requesting a valid
# storage service.
ERR_HTM = "<div>There is no data for this chart, please check you are looking at a valid storage service.</div>"

CSV_HEADERS = [
    fields.FIELD_AIP_UUID,
    fields.FIELD_AIP_NAME,
    fields.FIELD_DATE_START,
    fields.FIELD_DATE_END,
    fields.FIELD_USER,
    fields.FIELD_DURATION,
]


def get_table_data(ingests):
    """Format the data needed for an ingest log table and augment it
    where needed.
    """
    transfer_count = 0
    for ingest in ingests[fields.FIELD_INGESTS]:
        transfer_count = transfer_count + 1
        start_date = ingest[fields.FIELD_DATE_START]
        end_date = ingest[fields.FIELD_DATE_END]
        start_date_obj = _simplify_datetime(start_date)
        end_date_obj = _simplify_datetime(end_date)
        ingest[fields.FIELD_DURATION] = end_date_obj - start_date_obj
    ingests[TRANSFER_COUNT] = transfer_count
    return ingests


@reporter.route("/ingest_log_tabular/", methods=["GET"])
def ingest_log_tabular():
    """Return the information needed to present an ingest gantt chart."""
    storage_service_id = request.args.get(request_params.STORAGE_SERVICE_ID)
    storage_location_id = request.args.get(request_params.STORAGE_LOCATION_ID)
    start_date = parse_datetime_bound(request.args.get(request_params.START_DATE))
    end_date = parse_datetime_bound(
        request.args.get(request_params.END_DATE), upper=True
    )
    csv = parse_bool(request.args.get(request_params.CSV), default=False)

    ingests = report_data.agents_transfers(
        storage_service_id, start_date, end_date, storage_location_id
    )
    ingests = get_table_data(ingests)

    if csv:
        filename = "user_ingest_log.csv"
        headers = translate_headers(CSV_HEADERS)
        return download_csv(headers, ingests[fields.FIELD_INGESTS], filename)

    return render_template(
        "report_ingest_log_tabular.html",
        storage_service_name=ingests.get(fields.FIELD_STORAGE_NAME),
        storage_location_description=ingests.get(fields.FIELD_STORAGE_LOCATION),
        number_of_transfers=ingests[TRANSFER_COUNT],
        data=ingests[fields.FIELD_INGESTS],
        start_date=start_date,
        end_date=get_display_end_date(end_date),
    )


def get_figure_html(ingests):
    """Augment the ingest log data and append the Gantt's HTML source
    for printing in the report template.
    """
    pd_list = []
    transfer_count = 0
    for ingest in ingests[fields.FIELD_INGESTS]:
        transfer_count = transfer_count + 1
        pd_list.append(
            dict(
                User=ingest[fields.FIELD_USER],
                Start=ingest[fields.FIELD_DATE_START],
                Finish=ingest[fields.FIELD_DATE_END],
            )
        )
    if not pd_list:
        ingests[FIGURE_HTML] = ERR_HTM
        ingests[TRANSFER_COUNT] = transfer_count
        return ingests
    df = pd.DataFrame(pd_list)
    fig = px.timeline(df, x_start="Start", x_end="Finish", y="User")
    fig.update_yaxes(autorange="reversed")
    ingests[FIGURE_HTML] = fig.to_html(full_html=False)
    ingests[TRANSFER_COUNT] = transfer_count
    return ingests


@reporter.route("/ingest_log_gantt/", methods=["GET"])
def ingest_log():
    """Return the information needed to present an ingest gantt chart."""
    storage_service_id = request.args.get(request_params.STORAGE_SERVICE_ID)
    storage_location_id = request.args.get(request_params.STORAGE_LOCATION_ID)
    start_date = parse_datetime_bound(request.args.get(request_params.START_DATE))
    end_date = parse_datetime_bound(
        request.args.get(request_params.END_DATE), upper=True
    )
    ingests = report_data.agents_transfers(
        storage_service_id, start_date, end_date, storage_location_id
    )
    ingests = get_figure_html(ingests)
    return render_template(
        "report_ingest_log_gantt.html",
        storage_service_name=ingests.get(fields.FIELD_STORAGE_NAME),
        storage_location_description=ingests.get(fields.FIELD_STORAGE_LOCATION),
        number_of_transfers=ingests[TRANSFER_COUNT],
        plot=ingests[FIGURE_HTML],
        start_date=start_date,
        end_date=get_display_end_date(end_date),
    )



import pytest
from flask import current_app

EXPECTED_CSV_ORIGINAL = b"AIP Name,UUID,Count,Size,Size (bytes)\r\nTest AIP,111111111111-1111-1111-11111111,1,1.0 kB,1000\r\n"
EXPECTED_CSV_PRESERVATION = b"AIP Name,UUID,Count,Size,Size (bytes)\r\nTest AIP,111111111111-1111-1111-11111111,1,2.0 kB,2000\r\n"


@pytest.mark.parametrize(
    "original_files",
    [
        # Original files.
        ("True"),
        # Preservation files.
        ("False"),
    ],
)
def test_aips_by_file_format(app_with_populated_files, original_files):
    """Test that report template renders."""
    with current_app.test_client() as test_client:
        response = test_client.get(
            f"/reporter/aips_by_file_format/?amss_id=1&file_format=Tagged Image File Format&original_files={original_files}"
        )
        assert response.status_code == 200


@pytest.mark.parametrize(
    "original_files,expected_csv",
    [
        # Original files.
        ("True", EXPECTED_CSV_ORIGINAL),
        # Preservation files.
        ("False", EXPECTED_CSV_PRESERVATION),
    ],
)
def test_aips_by_file_format_csv(
    app_with_populated_files, original_files, expected_csv
):
    """Test CSV export."""
    with current_app.test_client() as test_client:
        response = test_client.get(
            f"/reporter/aips_by_file_format/?amss_id=1&file_format=Tagged Image File Format&original_files={original_files}&csv=True"
        )
        assert response.status_code == 200
        assert (
            response.headers["Content-Disposition"]
            == "attachment; filename=aips_by_file_format_Tagged Image File Format.csv"
        )
        assert response.mimetype == "text/csv"
        assert response.data == expected_csv



"""Field name constants."""

FIELD_AIP_ID = "AIPID"
FIELD_AIP_NAME = "AIPName"
FIELD_AIP_SIZE = "AIPSize"
FIELD_AIP_UUID = "AIPUUID"
FIELD_AIPS = "AIPs"
FIELD_ALL_AIPS = "AllAIPs"

FIELD_COUNT = "Count"
FIELD_CREATED_DATE = "CreatedDate"
FIELD_CURRENT_LOCATION = "CurrentLocation"

FIELD_DATE_START = "IngestStartDate"
FIELD_DATE_END = "IngestFinishDate"
FIELD_DERIVATIVE_COUNT = "DerivativeCount"
FIELD_DERIVATIVE_FILE = "DerivativeFile"
FIELD_DERIVATIVE_FORMAT = "DerivativeFormat"
FIELD_DERIVATIVE_UUID = "DerivativeUUID"
FIELD_DURATION = "Duration"

FIELD_EVENT = "Event"

FIELD_FILES = "Files"
FIELD_FILE_COUNT = "FileCount"
FIELD_FILE_TYPE = "FileType"
FIELD_FILENAME = "Filename"
FIELD_FORMAT = "Format"
FIELD_FORMATS = "Formats"
FIELD_FORMAT_VERSIONS = "FormatVersions"

FIELD_ID = "ID"
FIELD_INGESTS = "Ingests"
FIELD_INGEST_START_DATE = "IngestStartDate"
FIELD_INGEST_FINISH_DATE = "IngestFinishDate"

FIELD_LOCATIONS = "Locations"
FIELD_LOCATIONS_USAGE_OVER_TIME = "LocationsUsageOverTime"

FIELD_NAME = "Name"

FIELD_ORIGINAL_FILE = "OriginalFile"
FIELD_ORIGINAL_FORMAT = "OriginalFormat"
FIELD_ORIGINAL_NAME = "OriginalName"
FIELD_ORIGINAL_PUID = "OriginalPUID"
FIELD_ORIGINAL_UUID = "OriginalUUID"
FIELD_ORIGINAL_VERSION = "OriginalVersion"

FIELD_PUID = "PUID"

FIELD_RELATED_PAIRING = "RelatedPairing"

FIELD_SIZE = "Size"
FIELD_SIZE_BYTES = "SizeBytes"
FIELD_STORAGE_LOCATION = "StorageLocation"
FIELD_STORAGE_NAME = "StorageName"

FIELD_TRANSFER_NAME = "TransferName"

FIELD_USER = "User"
FIELD_UUID = "UUID"

FIELD_VERSION = "Version"



import enum
import re
from datetime import date
from datetime import datetime

from sqlalchemy.dialects import mysql

from AIPscan import db

UUID_REGEX = "[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}"


class package_tasks(db.Model):
    __tablename__ = "package_tasks"
    package_task_id = db.Column(db.String(36), primary_key=True)
    workflow_coordinator_id = db.Column(db.String(36))


class get_mets_tasks(db.Model):
    __tablename__ = "get_mets_tasks"
    get_mets_task_id = db.Column(db.String(36), primary_key=True)
    workflow_coordinator_id = db.Column(db.String(36))
    package_uuid = db.Column(db.String(36))
    status = db.Column(db.String(36))


class index_tasks(db.Model):
    __tablename__ = "index_tasks"
    index_task_id = db.Column(db.String(36), primary_key=True)
    fetch_job_id = db.Column(
        db.Integer(), db.ForeignKey("fetch_job.id"), nullable=False
    )
    indexing_start = db.Column(db.DateTime())
    indexing_progress = db.Column(db.String(255))
    indexing_end = db.Column(db.DateTime())


class StorageService(db.Model):
    __tablename__ = "storage_service"
    id = db.Column(db.Integer(), primary_key=True)
    name = db.Column(db.String(255), index=True, unique=True)
    url = db.Column(db.String(255))
    user_name = db.Column(db.String(255))
    api_key = db.Column(db.String(255))
    download_limit = db.Column(db.Integer())
    download_offset = db.Column(db.Integer())
    default = db.Column(db.Boolean)
    fetch_jobs = db.relationship(
        "FetchJob", cascade="all,delete", backref="storage_service", lazy=True
    )
    storage_locations = db.relationship(
        "StorageLocation", cascade="all,delete", backref="storage_service", lazy=True
    )

    def __init__(
        self, name, url, user_name, api_key, download_limit, download_offset, default
    ):
        self.name = name
        self.url = url
        self.user_name = user_name
        self.api_key = api_key
        self.download_limit = download_limit
        self.download_offset = download_offset
        self.default = default

    def __repr__(self):
        return f"<Storage Service '{self.name}'>"

    @property
    def earliest_aip_created(self):
        results = (
            db.session.query(AIP.create_date).order_by(AIP.create_date.asc()).first()
        )
        try:
            return results[0]
        except TypeError:
            return date.today()

    @property
    def unique_file_formats(self):
        return (
            db.session.query(File.file_format.distinct().label("name"))
            .join(AIP)
            .join(StorageService)
            .filter(StorageService.id == self.id)
            .order_by(File.file_format)
        )

    @property
    def unique_original_file_formats(self):
        formats = self.unique_file_formats
        original_formats = formats.filter(File.file_type == FileType.original)
        return [format_.name for format_ in original_formats]

    @property
    def unique_preservation_file_formats(self):
        formats = self.unique_file_formats
        preservation_formats = formats.filter(File.file_type == FileType.preservation)
        return [format_.name for format_ in preservation_formats]

    @property
    def unique_puids(self):
        return (
            db.session.query(File.puid.distinct().label("puid"))
            .join(AIP)
            .join(StorageService)
            .filter(StorageService.id == self.id)
            .order_by(File.puid)
        )

    @property
    def unique_original_puids(self):
        puids = self.unique_puids
        original_puids = puids.filter(File.file_type == FileType.original)
        return [puid.puid for puid in original_puids if puid.puid is not None]

    @property
    def unique_preservation_puids(self):
        puids = self.unique_puids
        preservation_puids = puids.filter(File.file_type == FileType.preservation)
        return [puid.puid for puid in preservation_puids if puid.puid is not None]


class StorageLocation(db.Model):
    __tablename__ = "storage_location"
    id = db.Column(db.Integer(), primary_key=True)
    current_location = db.Column(db.String(255), unique=True, index=True)
    description = db.Column(db.String(255))
    storage_service_id = db.Column(
        db.Integer(), db.ForeignKey("storage_service.id"), nullable=False
    )
    aips = db.relationship(
        "AIP", cascade="all,delete", backref="storage_location", lazy=True
    )

    def __init__(self, current_location, description, storage_service_id):
        self.current_location = current_location
        self.description = description
        self.storage_service_id = storage_service_id

    def __repr__(self):
        return f"<StorageLocation '{self.current_location}'>"

    @property
    def uuid(self):
        match = re.search(UUID_REGEX, self.current_location)
        if match:
            return match.group(0)
        return None

    @property
    def unique_file_formats(self):
        return (
            db.session.query(File.file_format.distinct().label("name"))
            .join(AIP)
            .join(StorageLocation)
            .filter(StorageLocation.id == self.id)
            .order_by(File.file_format)
        )

    @property
    def unique_original_file_formats(self):
        formats = self.unique_file_formats
        original_formats = formats.filter(File.file_type == FileType.original)
        return [format_.name for format_ in original_formats]

    @property
    def unique_preservation_file_formats(self):
        formats = self.unique_file_formats
        preservation_formats = formats.filter(File.file_type == FileType.preservation)
        return [format_.name for format_ in preservation_formats]

    @property
    def unique_puids(self):
        return (
            db.session.query(File.puid.distinct().label("puid"))
            .join(AIP)
            .join(StorageLocation)
            .filter(StorageLocation.id == self.id)
            .order_by(File.puid)
        )

    @property
    def unique_original_puids(self):
        puids = self.unique_puids
        original_puids = puids.filter(File.file_type == FileType.original)
        return [puid.puid for puid in original_puids if puid.puid is not None]

    @property
    def unique_preservation_puids(self):
        puids = self.unique_puids
        preservation_puids = puids.filter(File.file_type == FileType.preservation)
        return [puid.puid for puid in preservation_puids if puid.puid is not None]

    def aip_count(self, start_date=None, end_date=None):
        """Return count of AIPs in this location."""
        DEFAULT = 0
        if not start_date:
            start_date = datetime.min
        if not end_date:
            end_date = datetime.max
        results = (
            db.session.query(db.func.count(AIP.id))
            .join(StorageLocation)
            .filter(AIP.storage_location_id == self.id)
            .filter(AIP.create_date >= start_date)
            .filter(AIP.create_date < end_date)
            .first()
        )
        try:
            return results[0]
        except (IndexError, TypeError):
            return DEFAULT

    def aip_total_size(self, start_date=None, end_date=None):
        """Return size in bytes of all AIPs in this location."""
        DEFAULT = 0
        if not start_date:
            start_date = datetime.min
        if not end_date:
            end_date = datetime.max
        results = (
            db.session.query(db.func.sum(File.size))
            .join(AIP)
            .join(StorageLocation)
            .filter(AIP.storage_location_id == self.id)
            .filter(AIP.create_date >= start_date)
            .filter(AIP.create_date < end_date)
            .first()
        )
        if results[0]:
            try:
                return results[0]
            except (IndexError, TypeError):
                pass
        return DEFAULT

    def file_count(self, start_date=None, end_date=None, originals=True):
        """Return number of files in this location. Defaults to originals only."""
        if not start_date:
            start_date = datetime.min
        if not end_date:
            end_date = datetime.max
        files = (
            db.session.query(File)
            .join(AIP)
            .join(StorageLocation)
            .filter(AIP.storage_location_id == self.id)
            .filter(AIP.create_date >= start_date)
            .filter(AIP.create_date < end_date)
        )
        if originals:
            files = files.filter(File.file_type == FileType.original)
        return files.count()


class FetchJob(db.Model):
    __tablename__ = "fetch_job"
    id = db.Column(db.Integer(), primary_key=True)
    total_packages = db.Column(db.Integer())
    total_aips = db.Column(db.Integer())
    total_dips = db.Column(db.Integer())
    total_sips = db.Column(db.Integer())
    total_replicas = db.Column(db.Integer())
    total_deleted_aips = db.Column(db.Integer())
    download_start = db.Column(db.DateTime())
    download_end = db.Column(db.DateTime())
    download_directory = db.Column(db.String(255))
    storage_service_id = db.Column(
        db.Integer(), db.ForeignKey("storage_service.id"), nullable=False
    )
    aips = db.relationship("AIP", cascade="all,delete", backref="fetch_job", lazy=True)
    index_tasks = db.relationship(
        "index_tasks", cascade="all,delete", backref="fetch_job", lazy=True
    )

    def __init__(
        self,
        total_packages,
        total_aips,
        total_deleted_aips,
        download_start,
        download_end,
        download_directory,
        storage_service_id,
    ):
        self.total_packages = total_packages
        self.total_aips = total_aips
        self.total_deleted_aips = total_deleted_aips
        self.download_start = download_start
        self.download_end = download_end
        self.download_directory = download_directory
        self.storage_service_id = storage_service_id

    def __repr__(self):
        return f"<Fetch Job '{self.download_start}'>"


class Pipeline(db.Model):
    __tablename__ = "pipeline"
    id = db.Column(db.Integer(), primary_key=True)
    origin_pipeline = db.Column(db.String(255), unique=True, index=True)
    dashboard_url = db.Column(db.Text())
    aips = db.relationship(
        "AIP", cascade="all,delete", backref="origin_pipeline", lazy=True
    )

    def __init__(self, origin_pipeline, dashboard_url):
        self.origin_pipeline = origin_pipeline
        self.dashboard_url = dashboard_url

    def __repr__(self):
        return f"<Pipeline '{self.origin_pipeline}'>"

    @property
    def uuid(self):
        match = re.search(UUID_REGEX, self.origin_pipeline)
        if match:
            return match.group(0)
        return None


class AIP(db.Model):
    __tablename__ = "aip"
    id = db.Column(db.Integer(), primary_key=True)
    uuid = db.Column(db.String(255), index=True)
    transfer_name = db.Column(db.String(255))
    create_date = db.Column(db.DateTime(), index=True)
    mets_sha256 = db.Column(db.String(64))
    size = db.Column(db.BigInteger())
    storage_service_id = db.Column(
        db.Integer(), db.ForeignKey("storage_service.id"), nullable=False
    )
    storage_location_id = db.Column(
        db.Integer(), db.ForeignKey("storage_location.id"), nullable=False
    )
    fetch_job_id = db.Column(
        db.Integer(), db.ForeignKey("fetch_job.id"), nullable=False
    )
    origin_pipeline_id = db.Column(
        db.Integer(), db.ForeignKey("pipeline.id"), nullable=False
    )
    files = db.relationship("File", cascade="all,delete", backref="aip", lazy=True)

    def __init__(
        self,
        uuid,
        transfer_name,
        create_date,
        mets_sha256,
        size,
        storage_service_id,
        storage_location_id,
        fetch_job_id,
        origin_pipeline_id,
    ):
        self.uuid = uuid
        self.transfer_name = transfer_name
        self.create_date = create_date
        self.mets_sha256 = mets_sha256
        self.size = size
        self.storage_service_id = storage_service_id
        self.storage_location_id = storage_location_id
        self.fetch_job_id = fetch_job_id
        self.origin_pipeline_id = origin_pipeline_id

    def __repr__(self):
        return f"<AIP '{self.transfer_name}'>"

    @property
    def original_file_count(self):
        return File.query.filter_by(aip_id=self.id, file_type=FileType.original).count()

    @property
    def preservation_file_count(self):
        return File.query.filter_by(
            aip_id=self.id, file_type=FileType.preservation
        ).count()


class FileType(enum.Enum):
    original = "original"
    preservation = "preservation"


class File(db.Model):
    __tablename__ = "file"
    id = db.Column(db.Integer(), primary_key=True)
    name = db.Column(db.String(255), index=True)
    filepath = db.Column(db.Text(), nullable=True)  # Accommodate long filepaths.
    uuid = db.Column(db.String(255), index=True)
    file_type = db.Column(db.Enum(FileType))
    size = db.Column(db.BigInteger())
    # Date created maps to PREMIS dateCreatedByApplication for original
    # files, which in practice is almost always date last modified, and
    # to normalization date for preservation files.
    date_created = db.Column(db.DateTime(), index=True)
    puid = db.Column(db.String(255), index=True)
    file_format = db.Column(db.String(255))
    format_version = db.Column(db.String(255))
    checksum_type = db.Column(db.String(255))
    checksum_value = db.Column(db.String(255), index=True)
    premis_object = db.Column(mysql.LONGTEXT)

    original_file_id = db.Column(db.Integer(), db.ForeignKey("file.id"))
    original_file = db.relationship(
        "File", remote_side=[id], backref=db.backref("derivatives")
    )

    aip_id = db.Column(db.Integer(), db.ForeignKey("aip.id"), nullable=False)
    events = db.relationship("Event", cascade="all,delete", backref="file", lazy=True)

    def __init__(
        self,
        name,
        filepath,
        uuid,
        size,
        date_created,
        file_format,
        checksum_type,
        checksum_value,
        aip_id,
        file_type=FileType.original,
        premis_object=None,
        format_version=None,
        puid=None,
        original_file_id=None,
    ):
        self.name = name
        self.filepath = filepath
        self.uuid = uuid
        self.file_type = file_type
        self.size = size
        self.date_created = date_created
        self.puid = puid
        self.file_format = file_format
        self.format_version = format_version
        self.checksum_type = checksum_type
        self.checksum_value = checksum_value
        self.premis_object = premis_object
        self.original_file_id = original_file_id
        self.aip_id = aip_id

    def __repr__(self):
        return f"<File '{self.id}' - '{self.name}'"

    def get_premis_xml_lines(self):
        """Return PREMIS object XML split into lines (list of str)."""
        value = self.premis_object
        if value is None:
            return []
        return str(value).splitlines()


EventAgent = db.Table(
    "event_agents",
    db.Column("event_id", db.Integer, db.ForeignKey("event.id")),
    db.Column("agent_id", db.Integer, db.ForeignKey("agent.id")),
)


class Event(db.Model):
    __tablename__ = "event"
    id = db.Column(db.Integer(), primary_key=True)
    type = db.Column(db.String(255), index=True)
    uuid = db.Column(db.String(255), index=True)
    date = db.Column(db.DateTime())
    detail = db.Column(db.String(255))
    outcome = db.Column(db.String(255))
    outcome_detail = db.Column(mysql.LONGTEXT)
    file_id = db.Column(db.Integer(), db.ForeignKey("file.id"), nullable=False)
    event_agents = db.relationship(
        "Agent", secondary=EventAgent, backref=db.backref("Event", lazy="dynamic")
    )

    def __init__(
        self, event_type, uuid, date, detail, outcome, outcome_detail, file_id
    ):
        self.type = event_type
        self.uuid = uuid
        self.date = date
        self.detail = detail
        self.outcome = outcome
        self.outcome_detail = outcome_detail
        self.file_id = file_id

    def __repr__(self):
        return f"<Event '{self.type}'>"


class Agent(db.Model):
    __tablename__ = "agent"
    id = db.Column(db.Integer(), primary_key=True)
    linking_type_value = db.Column(db.String(255), index=True)
    agent_type = db.Column(db.String(255), index=True)
    agent_value = db.Column(db.String(255), index=True)
    storage_service_id = db.Column(
        db.Integer(), db.ForeignKey("storage_service.id"), nullable=False
    )

    def __init__(self, linking_type_value, agent_type, agent_value, storage_service_id):
        self.linking_type_value = linking_type_value
        self.agent_type = agent_type
        self.agent_value = agent_value
        self.storage_service_id = storage_service_id

    def __repr__(self):
        return f"<Agent '{self.agent_type}: {self.agent_value}'>"



from flask_wtf import FlaskForm
from wtforms import BooleanField
from wtforms import StringField
from wtforms.validators import DataRequired
from wtforms.widgets import PasswordInput


class StorageServiceForm(FlaskForm):
    name = StringField("Name", validators=[DataRequired()])
    url = StringField("URL", validators=[DataRequired()])
    user_name = StringField("User name", validators=[DataRequired()])
    api_key = StringField(
        "API key", validators=[DataRequired()], widget=PasswordInput(hide_value=False)
    )
    download_limit = StringField("Download limit", default="20")
    download_offset = StringField("Download offset", default="0")
    default = BooleanField("Default")



"""Tests for custom Flask CLI commands."""

from AIPscan.models import StorageService


def _run_bootstrap(runner, **kwargs):
    args = ["storage-service-bootstrap"]
    for key, value in kwargs.items():
        if isinstance(value, bool):
            if value:
                args.append(f"--{key.replace('_', '-')}")
            continue
        args.extend([f"--{key.replace('_', '-')}", str(value)])
    return runner.invoke(args=args)


def test_storage_service_bootstrap_creates_storage_service(app_instance):
    runner = app_instance.test_cli_runner()

    result = _run_bootstrap(
        runner,
        name="Primary",
        url="https://storage.example.org",
        username="demo",
        api_key="abc123",
        default=True,
    )

    assert result.exit_code == 0

    storage_service = StorageService.query.filter_by(name="Primary").one()
    assert storage_service.url == "https://storage.example.org"
    assert storage_service.user_name == "demo"
    assert storage_service.api_key == "abc123"
    assert storage_service.default is True
    assert storage_service.download_limit == 20
    assert storage_service.download_offset == 0


def test_storage_service_bootstrap_is_idempotent(app_instance):
    runner = app_instance.test_cli_runner()

    _run_bootstrap(
        runner,
        name="Primary",
        url="https://storage.example.org",
        username="demo",
        api_key="abc123",
        default=True,
    )

    result = _run_bootstrap(
        runner,
        name="Primary",
        url="https://storage.example.org",
        username="demo",
        api_key="abc123",
    )

    assert result.exit_code == 0
    assert StorageService.query.count() == 1


def test_storage_service_bootstrap_updates_default_flag(app_instance):
    runner = app_instance.test_cli_runner()

    _run_bootstrap(
        runner,
        name="Primary",
        url="https://storage.example.org",
        username="demo",
        api_key="abc123",
    )

    _run_bootstrap(
        runner,
        name="Secondary",
        url="https://storage.example.net",
        username="demo",
        api_key="xyz789",
        default=True,
    )

    primary = StorageService.query.filter_by(name="Primary").one()
    secondary = StorageService.query.filter_by(name="Secondary").one()

    assert primary.default is False
    assert secondary.default is True


def test_storage_service_bootstrap_updates_existing_credentials(app_instance):
    runner = app_instance.test_cli_runner()

    _run_bootstrap(
        runner,
        name="Primary",
        url="https://storage.example.org",
        username="demo",
        api_key="abc123",
    )

    result = _run_bootstrap(
        runner,
        name="Primary",
        url="https://storage.example.com",
        username="demo2",
        api_key="xyz789",
    )

    assert result.exit_code == 0

    storage_service = StorageService.query.filter_by(name="Primary").one()
    assert storage_service.url == "https://storage.example.com"
    assert storage_service.user_name == "demo2"
    assert storage_service.api_key == "xyz789"



import pytest

from AIPscan.conftest import STORAGE_LOCATION_1_DESCRIPTION
from AIPscan.conftest import STORAGE_LOCATION_2_DESCRIPTION
from AIPscan.Data import data
from AIPscan.Data import fields
from AIPscan.Data.tests import MOCK_STORAGE_SERVICES
from AIPscan.helpers import parse_datetime_bound

DATE_BEFORE_AIP_1 = "2019-01-01"
DATE_AFTER_AIP_1 = "2020-01-02"
DATE_BEFORE_AIP_2 = "2020-05-30"
DATE_AFTER_AIP_2 = "2020-06-02"
DATE_BEFORE_AIP_3 = "2021-05-30"
DATE_AFTER_AIP_3 = "2021-06-01"


TEST_AIP_1_EXPECTED = {
    "AIPName": "TestAIP1",
    "CreatedDate": "2020-01-01 00:00:00",
    "Formats": {
        "fmt/43": {"Count": 1, "Name": "JPEG", "Version": "1.01"},
        "fmt/test-1": {"Count": 1, "Name": "txt", "Version": "0.0.0"},
    },
    "Size": 300,
    "UUID": "111111111111-1111-1111-11111111",
}

TEST_AIP_2_EXPECTED = {
    "AIPName": "TestAIP2",
    "CreatedDate": "2020-06-01 00:00:00",
    "Formats": {"fmt/44": {"Count": 1, "Name": "JPEG", "Version": "1.02"}},
    "Size": 1000,
    "UUID": "222222222222-2222-2222-22222222",
}

TEST_AIP_3_EXPECTED = {
    "AIPName": "TestAIP3",
    "CreatedDate": "2021-05-31 00:00:00",
    "Formats": {
        "fmt/test-1": {"Count": 2, "Name": "ACME File Format", "Version": "0.0.0"}
    },
    "Size": 2500,
    "UUID": "333333333333-3333-3333-33333333",
}


@pytest.mark.parametrize(
    "storage_service_id, storage_service_name, storage_location_id, storage_location_description, start_date, end_date, expected_aips_count, expected_first_aip",
    [
        # Request for a Storage Service populated with two locations, each
        # containing AIPs and files.
        (
            1,
            "test storage service",
            None,
            None,
            DATE_BEFORE_AIP_1,
            DATE_AFTER_AIP_3,
            3,
            TEST_AIP_1_EXPECTED,
        ),
        # Request for a non-existent Storage Service.
        (4, None, None, None, DATE_BEFORE_AIP_1, DATE_BEFORE_AIP_3, 0, None),
        # Request for a None Storage Service.
        (None, None, None, None, DATE_BEFORE_AIP_1, DATE_BEFORE_AIP_3, 0, None),
        # Request with valid Storage Location.
        (
            1,
            "test storage service",
            1,
            STORAGE_LOCATION_1_DESCRIPTION,
            DATE_BEFORE_AIP_1,
            DATE_AFTER_AIP_3,
            2,
            TEST_AIP_1_EXPECTED,
        ),
        (
            1,
            "test storage service",
            2,
            STORAGE_LOCATION_2_DESCRIPTION,
            DATE_BEFORE_AIP_1,
            DATE_AFTER_AIP_3,
            1,
            TEST_AIP_3_EXPECTED,
        ),
        # Request for a non-existent Storage Location.
        (
            1,
            "test storage service",
            4,
            None,
            DATE_BEFORE_AIP_1,
            DATE_BEFORE_AIP_3,
            0,
            None,
        ),
        # Test date filtering.
        (
            1,
            "test storage service",
            None,
            None,
            DATE_AFTER_AIP_1,
            DATE_AFTER_AIP_2,
            1,
            TEST_AIP_2_EXPECTED,
        ),
        (
            1,
            "test storage service",
            None,
            None,
            DATE_BEFORE_AIP_1,
            DATE_BEFORE_AIP_2,
            1,
            TEST_AIP_1_EXPECTED,
        ),
        (
            1,
            "test storage service",
            None,
            None,
            DATE_BEFORE_AIP_1,
            DATE_BEFORE_AIP_3,
            2,
            TEST_AIP_1_EXPECTED,
        ),
        (
            1,
            "test storage service",
            None,
            None,
            DATE_AFTER_AIP_2,
            DATE_AFTER_AIP_3,
            1,
            TEST_AIP_3_EXPECTED,
        ),
    ],
)
def test_aip_contents_data(
    storage_locations,
    storage_service_id,
    storage_service_name,
    storage_location_id,
    storage_location_description,
    start_date,
    end_date,
    expected_aips_count,
    expected_first_aip,
):
    """Test response from report_data.storage_locations endpoint."""
    start_date = parse_datetime_bound(start_date)
    end_date = parse_datetime_bound(end_date, upper=True)

    report = data.aip_file_format_overview(
        storage_service_id=storage_service_id,
        start_date=start_date,
        end_date=end_date,
        storage_location_id=storage_location_id,
    )

    assert report[fields.FIELD_STORAGE_NAME] == storage_service_name
    assert report[fields.FIELD_STORAGE_LOCATION] == storage_location_description

    aips = report[fields.FIELD_AIPS]
    aips_count = len(aips)
    assert aips_count == expected_aips_count

    if aips_count < 1:
        return

    assert aips[0] == expected_first_aip


def test_storage_services(app_instance, mocker):
    """Test response from report_data.storage_services endpoint."""
    storage_services = mocker.patch("sqlalchemy.orm.query.Query.all")
    storage_services.return_value = MOCK_STORAGE_SERVICES

    report = data.storage_services()

    assert report == {
        "storage_services": [
            {"id": 1, "name": "some name"},
            {"id": 2, "name": "another name"},
        ]
    }



"""This module tests models."""

from datetime import date

import pytest

from AIPscan import db
from AIPscan import test_helpers
from AIPscan.conftest import AIP_CREATION_TIME
from AIPscan.conftest import JPEG_1_01_PUID
from AIPscan.conftest import JPEG_1_02_PUID
from AIPscan.conftest import PRESERVATION_FORMAT
from AIPscan.conftest import PRESERVATION_PUID
from AIPscan.conftest import STORAGE_LOCATION_1_CURRENT_LOCATION
from AIPscan.conftest import STORAGE_LOCATION_2_CURRENT_LOCATION
from AIPscan.conftest import TIFF_PUID
from AIPscan.helpers import parse_datetime_bound
from AIPscan.models import StorageLocation
from AIPscan.models import StorageService

VALID_UUID = "3ce6fbcb-cdfc-4cca-97e4-d19a469ca043"
VALID_CURRENT_LOCATION = f"/api/v2/location/{VALID_UUID}/"
VALID_ORIGIN_PIPELINE = f"/api/v2/pipeline/{VALID_UUID}/"

INVALID_UUID = "not-a-uuid"
INVALID_CURRENT_LOCATION = f"/api/v2/location/{INVALID_UUID}/"
INVALID_ORIGIN_PIPELINE = f"/api/v2/pipeline/{INVALID_UUID}/"


def test_new_storage_service(app_with_populated_files):
    """Test that a new storage service is created."""
    storage_service = test_helpers.create_test_storage_service(name="storage")
    assert storage_service.name == "storage"
    assert storage_service.url == "storage"
    assert storage_service.user_name == "test"
    assert storage_service.api_key == "test"
    assert storage_service.download_limit == 0
    assert storage_service.download_offset == 10
    assert storage_service.default is True


def test_storage_service_earliest_aip_created(app_with_populated_files):
    """Test Storage Service earliest_aip_created property with AIPs."""
    storage_service = test_helpers.create_test_storage_service(name="storage2")
    assert storage_service.earliest_aip_created == AIP_CREATION_TIME


def test_storage_service_earliest_aip_created_with_no_aips(app_instance):
    """Test Storage Service earliest_aip_created property without AIPs"""
    storage_service = test_helpers.create_test_storage_service(name="storage3")
    assert storage_service.earliest_aip_created == date.today()


@pytest.mark.parametrize(
    "storage_service_id, original_formats, preservation_formats",
    [
        # Test with populated Storage Service.
        (
            1,
            ["ACME File Format", "JPEG", "txt", "yet another format"],
            ["Tagged Image File Format", PRESERVATION_FORMAT],
        ),
        # Test with new Storage Service (no files).
        (None, [], []),
    ],
)
def test_storage_service_unique_file_formats(
    storage_locations, storage_service_id, original_formats, preservation_formats
):
    """Test Storage Service unique file format properties."""
    if storage_service_id:
        storage_service = db.session.get(StorageService, storage_service_id)
    else:
        storage_service = test_helpers.create_test_storage_service(
            name="empty storage service"
        )

    assert storage_service.unique_original_file_formats == original_formats
    assert storage_service.unique_preservation_file_formats == preservation_formats


@pytest.mark.parametrize(
    "storage_service_id, original_puids, preservation_puids",
    [
        # Test with populated Storage Service.
        (
            1,
            [JPEG_1_01_PUID, JPEG_1_02_PUID, "fmt/test-1"],
            [PRESERVATION_PUID, TIFF_PUID],
        ),
        # Test with new Storage Service (no files).
        (None, [], []),
    ],
)
def test_storage_service_unique_puids(
    storage_locations, storage_service_id, original_puids, preservation_puids
):
    """Test Storage Service unique PUID properties."""
    if storage_service_id:
        storage_service = db.session.get(StorageService, storage_service_id)
    else:
        storage_service = test_helpers.create_test_storage_service(
            name="empty storage service"
        )

    assert storage_service.unique_original_puids == original_puids
    assert storage_service.unique_preservation_puids == preservation_puids


@pytest.mark.parametrize(
    "current_location, expected_uuid",
    [
        # Test that UUID is returned for valid current location.
        (VALID_CURRENT_LOCATION, VALID_UUID),
        # Test that None is returned for invalid current location.
        (INVALID_CURRENT_LOCATION, None),
    ],
)
def test_storage_location_uuid(app_instance, current_location, expected_uuid):
    """Test Storage Location uuid property."""
    storage_location = test_helpers.create_test_storage_location(
        current_location=current_location
    )
    assert storage_location.uuid == expected_uuid


@pytest.mark.parametrize(
    "current_location, start_date, end_date, expected_aip_count",
    [
        # Test Storage Location populated with AIPs.
        (STORAGE_LOCATION_1_CURRENT_LOCATION, None, None, 2),
        # Test date parameters in populated Storage Location.
        (STORAGE_LOCATION_1_CURRENT_LOCATION, "2020-01-02", None, 1),
        (STORAGE_LOCATION_1_CURRENT_LOCATION, None, "2020-05-31", 1),
        (STORAGE_LOCATION_1_CURRENT_LOCATION, "2020-01-02", "2020-05-31", 0),
        (STORAGE_LOCATION_1_CURRENT_LOCATION, "2020-01-01", "2020-06-01", 2),
        # Test new Storage Location (no AIPs).
        (None, None, None, 0),
    ],
)
def test_storage_location_aip_count(
    storage_locations, current_location, start_date, end_date, expected_aip_count
):
    """Test Storage Location aip_count property."""
    if current_location:
        storage_location = StorageLocation.query.filter_by(
            current_location=current_location
        ).first()
    else:
        storage_location = test_helpers.create_test_storage_location()
    start_date = parse_datetime_bound(start_date)
    end_date = parse_datetime_bound(end_date, upper=True)
    assert storage_location.aip_count(start_date, end_date) == expected_aip_count


@pytest.mark.parametrize(
    "current_location, start_date, end_date, expected_total_size",
    [
        # Test Storage Location populated with AIPs and files.
        (STORAGE_LOCATION_1_CURRENT_LOCATION, None, None, 1600),
        # Test data parameters in populated Storage Location.
        (STORAGE_LOCATION_1_CURRENT_LOCATION, "2020-01-02", None, 1000),
        (STORAGE_LOCATION_1_CURRENT_LOCATION, None, "2020-05-31", 600),
        (STORAGE_LOCATION_1_CURRENT_LOCATION, "2020-01-02", "2020-05-31", 0),
        (STORAGE_LOCATION_1_CURRENT_LOCATION, "2020-01-01", "2020-06-01", 1600),
        # Test new Storage Location (no AIPs or files).
        (None, None, None, 0),
    ],
)
def test_storage_location_aip_total_size(
    storage_locations, current_location, start_date, end_date, expected_total_size
):
    """Test Storage Location aip_total_size property."""
    if current_location:
        storage_location = StorageLocation.query.filter_by(
            current_location=current_location
        ).first()
    else:
        storage_location = test_helpers.create_test_storage_location()
    start_date = parse_datetime_bound(start_date)
    end_date = parse_datetime_bound(end_date, upper=True)
    assert storage_location.aip_total_size(start_date, end_date) == expected_total_size


@pytest.mark.parametrize(
    "current_location, start_date, end_date, originals_only, expected_file_count",
    [
        # Test Storage Location populated with AIPs.
        (STORAGE_LOCATION_1_CURRENT_LOCATION, None, None, True, 3),
        (STORAGE_LOCATION_1_CURRENT_LOCATION, None, None, False, 4),
        # Test date parameters in populated Storage Location.
        (STORAGE_LOCATION_1_CURRENT_LOCATION, "2020-01-02", None, True, 1),
        (STORAGE_LOCATION_1_CURRENT_LOCATION, "2020-01-02", None, False, 1),
        (STORAGE_LOCATION_1_CURRENT_LOCATION, None, "2020-05-31", True, 2),
        (STORAGE_LOCATION_1_CURRENT_LOCATION, None, "2020-05-31", False, 3),
        (STORAGE_LOCATION_1_CURRENT_LOCATION, "2020-01-02", "2020-05-31", True, 0),
        (STORAGE_LOCATION_1_CURRENT_LOCATION, "2020-01-02", "2020-05-31", False, 0),
        (STORAGE_LOCATION_1_CURRENT_LOCATION, "2020-01-01", "2020-06-01", True, 3),
        (STORAGE_LOCATION_1_CURRENT_LOCATION, "2020-01-01", "2020-06-01", False, 4),
        # Test new Storage Location (no AIPs).
        (None, None, None, True, 0),
        (None, None, None, False, 0),
    ],
)
def test_storage_location_file_count(
    storage_locations,
    current_location,
    start_date,
    end_date,
    originals_only,
    expected_file_count,
):
    """Test Storage Location file_count property."""
    if current_location:
        storage_location = StorageLocation.query.filter_by(
            current_location=current_location
        ).first()
    else:
        storage_location = test_helpers.create_test_storage_location()
    start_date = parse_datetime_bound(start_date)
    end_date = parse_datetime_bound(end_date, upper=True)
    assert (
        storage_location.file_count(start_date, end_date, originals=originals_only)
        == expected_file_count
    )


@pytest.mark.parametrize(
    "current_location, original_formats, preservation_formats",
    [
        # Test with first Storage Location.
        (
            STORAGE_LOCATION_1_CURRENT_LOCATION,
            ["JPEG", "txt"],
            ["Tagged Image File Format"],
        ),
        # Test with second Storage Location.
        (
            STORAGE_LOCATION_2_CURRENT_LOCATION,
            ["ACME File Format", "yet another format"],
            [PRESERVATION_FORMAT],
        ),
        # Test with new Storage Location (no files).
        (None, [], []),
    ],
)
def test_storage_location_unique_file_formats(
    storage_locations, current_location, original_formats, preservation_formats
):
    """Test Storage Location unique file format properties."""
    if current_location:
        storage_location = StorageLocation.query.filter_by(
            current_location=current_location
        ).first()
    else:
        storage_location = test_helpers.create_test_storage_location()

    assert storage_location.unique_original_file_formats == original_formats
    assert storage_location.unique_preservation_file_formats == preservation_formats


@pytest.mark.parametrize(
    "current_location, original_puids, preservation_puids",
    [
        # Test with first Storage Location.
        (
            STORAGE_LOCATION_1_CURRENT_LOCATION,
            [JPEG_1_01_PUID, JPEG_1_02_PUID, "fmt/test-1"],
            [TIFF_PUID],
        ),
        # Test with second Storage Location.
        (STORAGE_LOCATION_2_CURRENT_LOCATION, ["fmt/test-1"], [PRESERVATION_PUID]),
        # Test with new Storage Location (no files).
        (None, [], []),
    ],
)
def test_storage_location_unique_puids(
    storage_locations, current_location, original_puids, preservation_puids
):
    """Test Storage Location unique PUID properties."""
    if current_location:
        storage_location = StorageLocation.query.filter_by(
            current_location=current_location
        ).first()
    else:
        storage_location = test_helpers.create_test_storage_location()

    assert storage_location.unique_original_puids == original_puids
    assert storage_location.unique_preservation_puids == preservation_puids


@pytest.mark.parametrize(
    "origin_pipeline, expected_uuid",
    [
        # Test that UUID is returned for valid current location.
        (VALID_ORIGIN_PIPELINE, VALID_UUID),
        # Test that None is returned for invalid current location.
        (INVALID_ORIGIN_PIPELINE, None),
    ],
)
def test_pipeline_uuid(app_instance, origin_pipeline, expected_uuid):
    """Test Storage Location uuid property."""
    pipeline = test_helpers.create_test_pipeline(origin_pipeline=origin_pipeline)
    assert pipeline.uuid == expected_uuid



"""Tests for Flask application factory configuration.

Ensures that SECRET_KEY is properly validated in production and that
development/test environments have appropriate fallback behavior.
"""

import pytest

from AIPscan import create_app


@pytest.mark.parametrize(
    "config_name,secret_value,should_raise,expected_secret",
    [
        # Production requires valid SECRET_KEY.
        ("default", None, True, None),
        ("default", "", True, None),
        ("default", "   ", True, None),
        ("default", "super-secret", False, "super-secret"),
        # Dev and test use fallback.
        ("dev", None, False, "dev"),
        ("test", None, False, "dev"),
    ],
    ids=[
        "prod-rejects-missing-key",
        "prod-rejects-empty-key",
        "prod-rejects-whitespace-key",
        "prod-accepts-valid-key",
        "dev-provides-fallback-key",
        "test-provides-fallback-key",
    ],
)
def test_create_app_secret_key_handling(
    monkeypatch, config_name, secret_value, should_raise, expected_secret
):
    """Test SECRET_KEY validation across all config environments."""
    monkeypatch.delenv("SECRET_KEY", raising=False)
    if secret_value is not None:
        monkeypatch.setenv("SECRET_KEY", secret_value)

    if should_raise:
        with pytest.raises(RuntimeError, match="SECRET_KEY must be set"):
            create_app(config_name)
    else:
        app = create_app(config_name)
        assert app.config["SECRET_KEY"] == expected_secret



import subprocess

from hatchling.builders.hooks.plugin.interface import BuildHookInterface


class CustomBuildHook(BuildHookInterface):
    def initialize(self, version, build_data):
        """Run npm build steps so static assets are packaged.

        Build frontend assets only when producing wheels; the sdist already
        includes prebuilt files.
        """
        if self.target_name != "wheel":
            return

        self._run_npm("clean-install")
        self._run_npm("run", "build")

    def _run_npm(self, *args: str) -> None:
        subprocess.run(["npm", *args], check=True, cwd=self.root)



def get_possible_storage_locations(storage_service):
    """Return list of Storage Locations if storage service provided."""
    if storage_service is None:
        return []

    return storage_service.storage_locations



import logging
import os
import time
from logging.config import fileConfig

from alembic import context
from flask import current_app
from sqlalchemy.exc import OperationalError

# This is the Alembic Config object.
# It provides access to the values within the .ini file in use.
config = context.config

# Interpret the config file for Python logging.
# This line sets up the base loggers.
fileConfig(config.config_file_name)
logger = logging.getLogger("alembic.env")


def get_engine():
    try:
        # This works with Flask-SQLAlchemy<3 and Alchemical.
        return current_app.extensions["migrate"].db.get_engine()
    except (TypeError, AttributeError):
        # This works with Flask-SQLAlchemy>=3.
        return current_app.extensions["migrate"].db.engine


def get_engine_url():
    try:
        return get_engine().url.render_as_string(hide_password=False).replace("%", "%%")
    except AttributeError:
        return str(get_engine().url).replace("%", "%%")


# Add your model's MetaData object here for 'autogenerate' support.
# For example, import your model module.
# Set target_metadata = mymodel.Base.metadata.
config.set_main_option("sqlalchemy.url", get_engine_url())
target_db = current_app.extensions["migrate"].db

# Other configuration values required by env.py can be acquired via
# config.get_main_option("my_important_option").
# ... etc.


def get_metadata():
    if hasattr(target_db, "metadatas"):
        return target_db.metadatas[None]
    return target_db.metadata


def run_migrations_offline():
    """Run migrations in 'offline' mode.

    This configures the context with just a URL
    and not an Engine, though an Engine is acceptable
    here as well.  By skipping the Engine creation
    we don't even need a DBAPI to be available.

    Calls to context.execute() here emit the given string to the
    script output.

    """
    url = config.get_main_option("sqlalchemy.url")
    context.configure(url=url, target_metadata=get_metadata(), literal_binds=True)

    with context.begin_transaction():
        context.run_migrations()


def run_migrations_online():
    """Run migrations in 'online' mode.

    In this scenario we need to create an Engine
    and associate a connection with the context.

    """

    # This callback prevents Alembic from generating empty migrations when
    # there are no schema changes.
    # Reference: http://alembic.zzzcomputing.com/en/latest/cookbook.html
    def process_revision_directives(context, revision, directives):
        if getattr(config.cmd_opts, "autogenerate", False):
            script = directives[0]
            if script.upgrade_ops.is_empty():
                directives[:] = []
                logger.info("No changes in schema detected.")

    conf_args = current_app.extensions["migrate"].configure_args
    if conf_args.get("process_revision_directives") is None:
        conf_args["process_revision_directives"] = process_revision_directives

    connectable = get_engine()
    max_attempts = int(os.getenv("ALEMBIC_CONNECT_RETRIES", "10"))
    retry_delay = float(os.getenv("ALEMBIC_CONNECT_RETRY_DELAY", "3"))

    attempt = 1
    while True:
        try:
            with connectable.connect() as connection:
                context.configure(
                    connection=connection, target_metadata=get_metadata(), **conf_args
                )

                with context.begin_transaction():
                    context.run_migrations()
            break
        except OperationalError as exc:
            if max_attempts and attempt >= max_attempts:
                logger.error(
                    "Could not connect to the database after %s attempts.",
                    attempt,
                )
                raise exc
            logger.warning(
                "Database not ready (attempt %s/%s): %s. Retrying in %.1f seconds.",
                attempt,
                "unlimited" if max_attempts == 0 else max_attempts,
                exc,
                retry_delay,
            )
            time.sleep(retry_delay)
            attempt += 1


if context.is_offline_mode():
    run_migrations_offline()
else:
    run_migrations_online()



"""This module contains code related to Celery configuration."""

import os

from celery import Celery

DEFAULT_CELERY_RESULT_BACKEND = (
    "db+mysql+pymysql://aipscan:demo@127.0.0.1:3406/celery?charset=utf8mb4"
)
DEFAULT_CELERY_BROKER_URL = "amqp://guest@localhost//"

celery = Celery(
    "tasks",
    backend=os.getenv("CELERY_RESULT_BACKEND", DEFAULT_CELERY_RESULT_BACKEND),
    broker=os.getenv("CELERY_BROKER_URL", DEFAULT_CELERY_BROKER_URL),
    include=["AIPscan.Aggregator.tasks"],
)


def configure_celery(app):
    """Add Flask app context to celery.Task."""
    TaskBase = celery.Task

    class ContextTask(TaskBase):
        abstract = True

        def __call__(self, *args, **kwargs):
            with app.app_context():
                return TaskBase.__call__(self, *args, **kwargs)

    celery.Task = ContextTask
    celery.conf.result_backend = os.getenv(
        "CELERY_RESULT_BACKEND", DEFAULT_CELERY_RESULT_BACKEND
    )
    celery.conf.broker_url = os.getenv("CELERY_BROKER_URL", DEFAULT_CELERY_BROKER_URL)
    return celery



from flask import current_app

EXPECTED_CSV_CONTENTS = b"UUID,AIP Name,Created Date,Size,Size (bytes),Formats\r\n111111111111-1111-1111-11111111,Test AIP,2020-01-01 00:00:00,0 Bytes,0,fmt/43 (ACME File Format 0.0.0): 1 file|fmt/61 (ACME File Format 0.0.0): 1 file\r\n222222222222-2222-2222-22222222,Test AIP,2020-06-01 00:00:00,0 Bytes,0,x-fmt/111 (ACME File Format 0.0.0): 3 files|fmt/61 (ACME File Format 0.0.0): 2 files\r\n"


def test_aip_contents(aip_contents):
    """Test that report template renders."""
    with current_app.test_client() as test_client:
        response = test_client.get("/reporter/aip_contents/?amss_id=1")
        assert response.status_code == 200


def test_aip_contents_csv(aip_contents):
    """Test CSV export."""
    with current_app.test_client() as test_client:
        response = test_client.get("/reporter/aip_contents/?amss_id=1&csv=True")
        assert response.status_code == 200
        assert (
            response.headers["Content-Disposition"]
            == "attachment; filename=aip_contents.csv"
        )
        assert response.mimetype == "text/csv"
        assert response.data == EXPECTED_CSV_CONTENTS



import os
import time
from datetime import datetime

from celery.result import AsyncResult
from flask import Blueprint
from flask import abort
from flask import flash
from flask import jsonify
from flask import redirect
from flask import render_template
from flask import request
from flask import url_for

from AIPscan import db
from AIPscan import decorators
from AIPscan import typesense_helpers
from AIPscan.Aggregator import database_helpers
from AIPscan.Aggregator import tasks
from AIPscan.Aggregator.downloads import get_download_root
from AIPscan.Aggregator.forms import StorageServiceForm
from AIPscan.Aggregator.task_helpers import format_api_url_with_limit_offset
from AIPscan.Aggregator.task_helpers import get_packages_directory
from AIPscan.Aggregator.tasks import TaskError
from AIPscan.celery import celery

# Custom celery Models.
from AIPscan.models import FetchJob
from AIPscan.models import StorageService
from AIPscan.models import get_mets_tasks
from AIPscan.models import index_tasks
from AIPscan.models import package_tasks

aggregator = Blueprint("aggregator", __name__, template_folder="templates")


# PICTURAE TODO: Starting to see patterns shared across modules, e.g.
# date handling in the data module and in here. Let's bring those
# together in a helpful kind of way.
def _split_ms(date_string):
    """Remove microseconds from the given date string."""
    return str(date_string).split(".")[0]


def _format_date(date_string):
    """Format date to something nicer that can played back in reports"""
    DATE_FORMAT_FULL = "%Y-%m-%d %H:%M:%S"
    DATE_FORMAT_PARTIAL = "%Y-%m-%d"
    formatted_date = datetime.strptime(_split_ms(date_string), DATE_FORMAT_FULL)
    return formatted_date.strftime(DATE_FORMAT_PARTIAL)


def _test_storage_service_connection(storage_service):
    """Test Storage Service credentials.

    :param storage_service: StorageService instance

    :raises ConnectionError: if credentials are invalid
    """
    _, request_url_without_api_key, request_url = format_api_url_with_limit_offset(
        storage_service
    )

    try:
        _ = tasks.make_request(request_url, request_url_without_api_key)
    except TaskError as err:
        raise ConnectionError(str(err)) from err


@aggregator.route("/", methods=["GET"])
def ss_default():
    # load the default storage service
    storage_service = StorageService.query.filter_by(default=True).first()
    if storage_service is None:
        # no default is set, retrieve the first storage service
        storage_service = StorageService.query.first()
        # there are no storage services defined at all
        if storage_service is None:
            return redirect(url_for("aggregator.storage_services"))
    mets_fetch_jobs = FetchJob.query.filter_by(
        storage_service_id=storage_service.id
    ).all()
    return render_template(
        "storage_service.html",
        storage_service=storage_service,
        mets_fetch_jobs=mets_fetch_jobs,
    )


@aggregator.route("/storage_service/<storage_service_id>", methods=["GET"])
def storage_service(storage_service_id):
    storage_service = db.session.get(StorageService, storage_service_id)

    if storage_service is None:
        abort(404)

    mets_fetch_jobs = FetchJob.query.filter_by(
        storage_service_id=storage_service_id
    ).all()
    return render_template(
        "storage_service.html",
        storage_service=storage_service,
        mets_fetch_jobs=mets_fetch_jobs,
    )


@aggregator.route("/storage_services", methods=["GET"])
def storage_services():
    storage_services = StorageService.query.all()
    return render_template("storage_services.html", storage_services=storage_services)


@aggregator.route("/edit_storage_service/<storage_service_id>", methods=["GET", "POST"])
def edit_storage_service(storage_service_id):
    form = StorageServiceForm()
    storage_service = db.session.get(StorageService, storage_service_id)

    if storage_service is None:
        abort(404)

    if request.method == "GET":
        form.name.data = storage_service.name
        form.url.data = storage_service.url
        form.user_name.data = storage_service.user_name
        form.download_limit.data = storage_service.download_limit
        form.download_offset.data = storage_service.download_offset
        form.api_key.data = storage_service.api_key
        form.default.data = storage_service.default
    if form.validate_on_submit():
        storage_service.name = form.name.data
        storage_service.url = form.url.data
        storage_service.user_name = form.user_name.data
        storage_service.api_key = form.api_key.data
        storage_service.download_limit = form.download_limit.data
        storage_service.download_offset = form.download_offset.data
        if form.default.data is True:
            storage_services = StorageService.query.all()
            for ss in storage_services:
                ss.default = False
        storage_service.default = form.default.data
        db.session.commit()
        flash(f"Storage service {form.name.data} updated")
        return redirect(url_for("aggregator.storage_services"))
    return render_template(
        "edit_storage_service.html", title="Storage Service", form=form
    )


@aggregator.route("/new_storage_service", methods=["GET", "POST"])
def new_storage_service():
    form = StorageServiceForm()
    if form.validate_on_submit():
        ss = StorageService(
            name=form.name.data,
            url=form.url.data,
            user_name=form.user_name.data,
            api_key=form.api_key.data,
            download_limit=form.download_limit.data,
            download_offset=form.download_offset.data,
            default=form.default.data,
        )
        db.session.add(ss)
        db.session.commit()
        flash(f"New storage service {form.name.data} created")
        return redirect(url_for("aggregator.storage_services"))
    return render_template(
        "edit_storage_service.html", title="Storage Service", form=form
    )


@aggregator.route("/delete_storage_service/<storage_service_id>", methods=["GET"])
@decorators.confirm_required(
    StorageService,
    "storage_service_id",
    "Are you sure you'd like to delete this storage service?",
    "Delete",
    "aggregator.ss_default",
)
def delete_storage_service(storage_service_id):
    storage_service = db.session.get(StorageService, storage_service_id)

    if storage_service is None:
        abort(404)

    tasks.delete_storage_service.delay(storage_service_id)

    return render_template(
        "result_message.html",
        title="Storage Service",
        message=f"Storage service '{storage_service.name}' is being deleted",
    )


@aggregator.route("/new_fetch_job/<fetch_job_id>", methods=["POST"])
def new_fetch_job(fetch_job_id):
    """Fetch and process AIP METS files from Storage Service."""
    storage_service = db.session.get(StorageService, fetch_job_id)

    # Check Storage Service credentials and return 400 if invalid prior to
    # creating the Fetch Job and kicking off the Celery task.
    try:
        _test_storage_service_connection(storage_service)
    except ConnectionError:
        return jsonify({}), 400

    # create a subdirectory for the download job using a timestamp string as its name
    datetime_obj_start = datetime.now().replace(microsecond=0)
    timestamp_str = datetime_obj_start.strftime("%Y-%m-%d-%H-%M-%S")
    timestamp = datetime_obj_start.strftime("%Y-%m-%d %H:%M:%S")

    download_root = get_download_root()

    try:
        os.makedirs(
            os.path.join(download_root, timestamp_str, "packages"), exist_ok=True
        )
        os.makedirs(os.path.join(download_root, timestamp_str, "mets"), exist_ok=True)
    except PermissionError:
        return jsonify({"message": "Permissions error"}), 500

    # create a fetch_job record in the aipscan database
    fetch_job = database_helpers.create_fetch_job(
        datetime_obj_start, timestamp_str, storage_service.id
    )

    packages_directory = get_packages_directory(timestamp_str)

    # send the METS fetch job to a background job that will coordinate other workers
    task = tasks.workflow_coordinator.delay(
        timestamp_str, storage_service.id, fetch_job.id, packages_directory
    )

    fetch_job_id = fetch_job.id
    task_id = _wait_for_package_list_task_id(task.id)
    if task_id is None:
        return (
            jsonify(
                {
                    "message": "The background task failed to start in a timely manner. Please check the system logs or try again later."
                }
            ),
            500,
        )

    # Send response back to JavaScript function that was triggered by
    # the 'New Fetch Job' button.
    response = {"timestamp": timestamp, "taskId": task_id, "fetchJobId": fetch_job_id}
    return jsonify(response)


@aggregator.route("/delete_fetch_job/<fetch_job_id>", methods=["GET"])
@decorators.confirm_required(
    FetchJob,
    "fetch_job_id",
    "Are you sure you'd like to delete this fetch job?",
    "Delete",
    "aggregator.ss_default",
)
def delete_fetch_job(fetch_job_id):
    fetch_job = db.session.get(FetchJob, fetch_job_id)

    if fetch_job is None:
        abort(404)

    storage_service = db.session.get(StorageService, fetch_job.storage_service_id)
    tasks.delete_fetch_job.delay(fetch_job_id)

    flash(f"Fetch job {fetch_job.download_start} is being deleted")
    return redirect(
        url_for("aggregator.storage_service", storage_service_id=storage_service.id)
    )


@aggregator.route("/package_list_task_status/<taskid>")
def task_status(taskid):
    task = tasks.package_lists_request.AsyncResult(taskid, app=celery)
    if task.state == "PENDING":
        # job did not start yet
        response = {"state": task.state}
    elif task.state != "FAILURE":
        if task.state == "SUCCESS":
            obj = package_tasks.query.filter_by(package_task_id=task.id).first()
            coordinator_id = obj.workflow_coordinator_id
            response = {"state": task.state, "coordinatorId": coordinator_id}
        else:
            response = {"state": task.state, "message": task.info.get("message")}
    # CR Note: I am not sure we will ever meet this else, because
    # task.state will always be intercepted through != FAILUTE. IDK. Do
    # you read it the same way?
    else:
        # something went wrong in the background job
        response = {
            "state": task.state,
            "status": str(task.info),  # this is the exception raised
        }
    return jsonify(response)


@aggregator.route("/get_mets_task_status/<coordinatorid>")
def get_mets_task_status(coordinatorid):
    """Get mets task status"""
    totalAIPs = int(request.args.get("totalAIPs"))
    fetchJobId = int(request.args.get("fetchJobId"))
    mets_tasks = get_mets_tasks.query.filter_by(
        workflow_coordinator_id=coordinatorid, status=None
    ).all()
    response = []
    incomplete = 0
    for row in mets_tasks:
        task_id = row.get_mets_task_id
        package_uuid = row.package_uuid
        task_result = AsyncResult(id=task_id, app=celery)
        mets_task_status = task_result.state
        if mets_task_status is None:
            incomplete = incomplete + 1
            continue
        if (mets_task_status == "SUCCESS") or (mets_task_status[0] == "FAILURE"):
            totalAIPs += 1
            response.append(
                {
                    "state": mets_task_status,
                    "package": f"METS.{package_uuid}.xml",
                    "totalAIPs": totalAIPs,
                }
            )
            obj = get_mets_tasks.query.filter_by(get_mets_task_id=task_id).first()
            obj.status = mets_task_status
            db.session.commit()
    if len(mets_tasks) != 0:
        return jsonify(response)
    if incomplete > 0 and totalAIPs == 0:
        response = {"state": "PENDING"}
        return jsonify(response)
    downloadEnd = datetime.now().replace(microsecond=0)
    obj = FetchJob.query.filter_by(id=fetchJobId).first()
    start = obj.download_start
    downloadStart = _format_date(start)
    obj.download_end = downloadEnd
    db.session.commit()
    response = {"state": "COMPLETED"}
    flash(f"Fetch Job {downloadStart} completed")
    return jsonify(response)


@aggregator.route("/index_refresh/<fetch_job_id>")
def index_refresh(fetch_job_id):
    if not typesense_helpers.typesense_enabled():
        abort(422)

    tasks.start_index_task(fetch_job_id)

    response = {"state": "ACCEPTED"}
    return jsonify(response)


@aggregator.route("/indexing_status/<fetch_job_id>")
def index_status(fetch_job_id):
    if not typesense_helpers.typesense_enabled():
        abort(422)

    # Get status from DB
    obj = index_tasks.query.filter_by(fetch_job_id=fetch_job_id).first()

    response = {}

    if obj.indexing_end is None:
        response["progress"] = obj.indexing_progress
        response["state"] = "PENDING"
    else:
        response = {"state": "COMPLETED"}

    return jsonify(response)


def _wait_for_package_list_task_id(
    coordinator_task_id, timeout_seconds=60, sleep_seconds=0.2
):
    """Wait briefly for the child package-list task id created by the coordinator.

    The coordinator worker enqueues a child task and writes a mapping row
    (workflow_coordinator_id -> package_task_id). Due to DB transaction
    isolation (e.g. MySQL REPEATABLE READ) the current session may not see
    that worker's commit, so we remove the session between polls to get fresh
    reads.

    This spin-wait is a pragmatic shortcut. Long-term, avoid server-side
    polling: return immediately with the coordinator id and have the client
    poll a dedicated endpoint until the child id appears.

    Returns the child task id string or None if not observed within timeout.
    """
    start_time = time.monotonic()
    while time.monotonic() - start_time < timeout_seconds:
        # End current transaction to avoid repeatable-read snapshot issues.
        db.session.remove()
        obj = package_tasks.query.filter_by(
            workflow_coordinator_id=coordinator_task_id
        ).first()
        if obj and obj.package_task_id:
            return obj.package_task_id
        time.sleep(sleep_seconds)
    return None



import json
import os
import shutil
from datetime import datetime

import requests
from celery.utils.log import get_task_logger

from AIPscan import db
from AIPscan import typesense_helpers
from AIPscan.Aggregator import database_helpers
from AIPscan.Aggregator.celery_helpers import with_db_session
from AIPscan.Aggregator.celery_helpers import write_celery_update
from AIPscan.Aggregator.mets_parse_helpers import METSError
from AIPscan.Aggregator.mets_parse_helpers import download_mets
from AIPscan.Aggregator.mets_parse_helpers import get_aip_original_name
from AIPscan.Aggregator.mets_parse_helpers import parse_mets_with_metsrw
from AIPscan.Aggregator.task_helpers import format_api_url_with_limit_offset
from AIPscan.Aggregator.task_helpers import parse_package_list_file
from AIPscan.Aggregator.task_helpers import process_package_object
from AIPscan.Aggregator.task_helpers import summarize_fetch_job_results
from AIPscan.celery import celery
from AIPscan.models import AIP
from AIPscan.models import Agent
from AIPscan.models import FetchJob
from AIPscan.models import StorageService
from AIPscan.models import get_mets_tasks
from AIPscan.models import index_tasks

logger = get_task_logger(__name__)


class TaskError(Exception):
    """Exception to call when there is a problem downloading from the
    storage service. The exception is known and asks for user
    intervention.
    """


def write_packages_json(count, packages, packages_directory):
    """Write package JSON to disk"""
    json_download_file = os.path.join(packages_directory, f"packages{count}.json")
    logger.info("Packages file is downloaded to '%s'", json_download_file)
    try:
        with open(json_download_file, "w") as json_file:
            json.dump(packages, json_file, indent=4)
    except json.JSONDecodeError:
        logger.error("Cannot decode JSON from %s", json_download_file)


def start_mets_task(
    package_uuid,
    aip_size,
    relative_path_to_mets,
    current_location,
    origin_pipeline,
    timestamp_str,
    package_list_no,
    storage_service_id,
    fetch_job_id,
    run_as_task=True,
):
    """Initiate a get_mets task worker and record the event in the
    celery database.
    """
    storage_service = db.session.get(StorageService, storage_service_id)
    storage_location = database_helpers.create_or_update_storage_location(
        current_location, storage_service
    )

    pipeline = database_helpers.create_or_update_pipeline(
        origin_pipeline, storage_service
    )

    args = [
        package_uuid,
        aip_size,
        relative_path_to_mets,
        timestamp_str,
        package_list_no,
        storage_service_id,
        storage_location.id,
        pipeline.id,
        fetch_job_id,
    ]

    if run_as_task:
        # Call worker to download and parse METS File.
        get_mets_task = get_mets.delay(*args)
        mets_task = get_mets_tasks(
            get_mets_task_id=get_mets_task.id,
            workflow_coordinator_id=workflow_coordinator.request.id,
            package_uuid=package_uuid,
            status=None,
        )
        db.session.add(mets_task)
        db.session.commit()
    else:
        # Execute immediately.
        get_mets.apply(args=args)


def delete_aip(uuid):
    logger.warning("Package deleted from SS: '%s'", uuid)

    deleted_aip = AIP.query.filter_by(uuid=uuid).first()

    if deleted_aip is not None:
        logger.warning("Deleting AIP: %s", uuid)
        database_helpers.delete_aip_object(deleted_aip)


@celery.task(bind=True)
@with_db_session
def workflow_coordinator(
    self, timestamp, storage_service_id, fetch_job_id, packages_directory
):
    logger.info("Packages directory is: %s", packages_directory)

    # Send package list request to a worker.
    package_lists_task = package_lists_request.delay(
        storage_service_id, timestamp, packages_directory
    )

    write_celery_update(package_lists_task, workflow_coordinator)

    # Wait for package lists task to finish downloading all package
    # lists.
    task = package_lists_request.AsyncResult(package_lists_task.id, app=celery)
    while True:
        if (task.state == "SUCCESS") or (task.state == "FAILURE"):
            break

    if isinstance(package_lists_task.info, TaskError):
        # Re-raise.
        raise (package_lists_task.info)

    total_package_lists = package_lists_task.info["totalPackageLists"]

    all_packages = []
    for package_list_no in range(1, total_package_lists + 1):
        json_file_path = os.path.join(
            packages_directory, f"packages{package_list_no}.json"
        )
        # Process packages and create a new worker to download and parse
        # each METS separately.
        packages = parse_package_list_file(json_file_path, logger, True)

        packages = process_packages(
            packages, storage_service_id, timestamp, package_list_no, fetch_job_id, True
        )

        all_packages = all_packages + packages

    total_packages_count = package_lists_task.info["totalPackages"]

    obj = database_helpers.update_fetch_job(
        fetch_job_id, all_packages, total_packages_count
    )

    logger.info("%s", summarize_fetch_job_results(obj))


def make_request(request_url, request_url_without_api_key):
    """Make our request to the storage service and return a valid
    response to our caller or raise a TaskError for celery.
    """
    response = requests.get(request_url)
    if response.status_code != requests.codes.ok:
        err = f"Check the URL and API details, cannot connect to: `{request_url_without_api_key}`"
        logger.error(err)
        raise TaskError(f"Bad response from server: {err}")
    try:
        packages = response.json()
    except json.JSONDecodeError as exc:
        msg = "Response is OK, but cannot decode JSON from server"
        logger.error(msg)
        raise TaskError(msg) from exc
    return packages


@celery.task(bind=True)
@with_db_session
def package_lists_request(self, storage_service_id, timestamp, packages_directory):
    """Request package lists from the storage service. Package lists
    will contain details of the AIPs that we want to download.
    """
    META = "meta"
    NEXT = "next"
    COUNT = "total_count"
    IN_PROGRESS = "IN PROGRESS"

    storage_service = db.session.get(StorageService, storage_service_id)

    (
        base_url,
        request_url_without_api_key,
        request_url,
    ) = format_api_url_with_limit_offset(storage_service)

    # First packages request.
    packages = make_request(request_url, request_url_without_api_key)
    packages_count = 1

    # Calculate how many package list files will be downloaded based on
    # total number of packages and the download limit
    total_packages = int(packages.get(META, {}).get(COUNT, 0))
    total_package_lists = int(total_packages / int(storage_service.download_limit)) + (
        total_packages % int(storage_service.download_limit) > 0
    )

    # There may be more packages to download to let's access those here.
    # TODO: `request_url_without_api_key` at this point will not be as
    # accurate. If we have more time, modify `format_api_url_with_limit_offset(...)`
    # to work with raw offset and limit data to make up for the fact
    # that an API key is plain-encoded in next_url.
    next_url = packages.get(META, {}).get(NEXT, None)
    write_packages_json(packages_count, packages, packages_directory)
    while next_url is not None:
        next_request = f"{base_url}{next_url}"
        next_packages = make_request(next_request, request_url_without_api_key)
        packages_count += 1
        write_packages_json(packages_count, next_packages, packages_directory)
        next_url = next_packages.get(META, {}).get(NEXT, None)
        self.update_state(
            state=IN_PROGRESS,
            meta={
                "message": f"Total packages: {total_packages} Total package lists: {total_package_lists}"
            },
        )

    return {
        "totalPackageLists": total_package_lists,
        "totalPackages": total_packages,
        "timestampStr": timestamp,
    }


def start_index_task(fetch_job_id):
    task = index_task.delay(fetch_job_id)

    index_task_obj = index_tasks(
        index_task_id=task.id, fetch_job_id=fetch_job_id, indexing_start=datetime.now()
    )
    db.session.add(index_task_obj)
    db.session.commit()


@celery.task()
@with_db_session
def index_task(fetch_job_id):
    # Update Typesense index
    typesense_helpers.initialize_index()

    index_task_obj = index_tasks.query.filter_by(fetch_job_id=fetch_job_id).first()

    for status in typesense_helpers.populate_index():
        if status["percent"] is not None:
            index_task_obj.indexing_progress = (
                f"Indexing {status['type']} data ({status['percent']}%)"
            )

            db.session.add(index_task_obj)
            db.session.commit()

    # Record indexing end time
    index_task_obj.indexing_end = datetime.now()

    db.session.add(index_task_obj)
    db.session.commit()


@celery.task()
@with_db_session
def get_mets(
    package_uuid,
    aip_size,
    relative_path_to_mets,
    timestamp_str,
    package_list_no,
    storage_service_id,
    storage_location_id,
    origin_pipeline_id,
    fetch_job_id,
    customlogger=None,
):
    """Request METS XML file from the storage service and parse.

    Download a METS file from an AIP that is stored in the storage
    service and then parse the results into the AIPscan database.

    This function relies on being able to use mets-reader-writer which
    is the primary object we will be passing about.

    TODO: Log METS errors.

    The "customlogger" argument allows an external logger to be specified when
    the task's logic is executed, using the task's "apply" method, by an
    external application like a batch script.
    """
    # Set logger
    tasklogger = logger
    if customlogger is not None:
        tasklogger = customlogger

    # Download METS file
    storage_service = db.session.get(StorageService, storage_service_id)
    download_file, mets_hash = download_mets(
        storage_service,
        package_uuid,
        relative_path_to_mets,
        timestamp_str,
        package_list_no,
    )
    mets_name = os.path.basename(download_file)

    # If METS file's hash matches an existing value, this is a duplicate of an
    # existing AIP and we can safely ignore it.
    matching_aip = AIP.query.filter_by(mets_sha256=mets_hash).first()
    if matching_aip is not None:
        tasklogger.info(
            f"Skipping METS file {mets_name} - identical to existing record"
        )
        try:
            os.remove(download_file)
        except OSError as err:
            tasklogger.warning(f"Unable to delete METS file: {err}")
        return

    tasklogger.info(f"Processing METS file {mets_name}")

    try:
        mets = parse_mets_with_metsrw(download_file)
    except METSError:
        # An error we need to log and report back to the user.
        return

    try:
        original_name = get_aip_original_name(mets)
    except METSError:
        # Some other error with the METS file that we might want to
        # log and act upon.
        original_name = package_uuid

    # Delete records of any previous versions of this AIP, which will shortly
    # be replaced by new records from the updated METS.
    previous_aips = AIP.query.filter_by(uuid=package_uuid).all()
    for previous_aip in previous_aips:
        tasklogger.info(
            f"Deleting record for AIP {package_uuid} to replace from newer METS"
        )
        database_helpers.delete_aip_object(previous_aip)

    aip = database_helpers.create_aip_object(
        package_uuid=package_uuid,
        transfer_name=original_name,
        create_date=mets.createdate,
        mets_sha256=mets_hash,
        size=aip_size,
        storage_service_id=storage_service_id,
        storage_location_id=storage_location_id,
        fetch_job_id=fetch_job_id,
        origin_pipeline_id=origin_pipeline_id,
    )

    database_helpers.process_aip_data(aip, mets)

    # Delete downloaded METS file.
    try:
        os.remove(download_file)
    except OSError as err:
        tasklogger.warning(f"Unable to delete METS file: {err}")


@celery.task()
@with_db_session
def delete_fetch_job(fetch_job_id):
    fetch_job = db.session.get(FetchJob, fetch_job_id)
    if os.path.exists(fetch_job.download_directory):
        shutil.rmtree(fetch_job.download_directory)
    db.session.delete(fetch_job)
    db.session.commit()


@celery.task()
@with_db_session
def delete_storage_service(storage_service_id):
    storage_service = db.session.get(StorageService, storage_service_id)
    mets_fetch_jobs = FetchJob.query.filter_by(
        storage_service_id=storage_service_id
    ).all()
    for mets_fetch_job in mets_fetch_jobs:
        if os.path.exists(mets_fetch_job.download_directory):
            shutil.rmtree(mets_fetch_job.download_directory)

    # Delete agents associated to the deleted storage service
    agents = Agent.query.filter_by(storage_service_id=storage_service_id).all()
    for agent in agents:
        logger.info("Deleting agent: %s", agent)
        db.session.delete(agent)

    db.session.delete(storage_service)
    db.session.commit()


def handle_deletion(package):
    if package.is_deleted():
        delete_aip(package.uuid)


def process_packages(
    packages,
    storage_service_id,
    timestamp_str,
    package_list_no,
    fetch_job_id,
    run_as_task=False,
    logger=None,
    start_item=None,
    end_item=None,
):
    """Parse packages documents from the storage service and initiate
    the load mets functions of AIPscan. Results are written to the
    database.
    """
    processed_packages = []

    package_count = 0
    for package_obj in packages.get("objects", []):
        package_count += 1

        package = process_package_object(package_obj)

        # Only process packages within paging window, if specified
        if start_item is None or (
            package_count >= start_item and package_count <= end_item
        ):
            # Calculate current item being processed
            if start_item is not None:
                current_item = start_item + len(processed_packages)

                if logger:
                    logger.info(
                        f"Processing {package.uuid} ({current_item} of {end_item})"
                    )

            processed_packages.append(package)
            handle_deletion(package)

            if not package.is_undeleted_aip():
                continue

            start_mets_task(
                package.uuid,
                package.size,
                package.get_relative_path(),
                package.current_location,
                package.origin_pipeline,
                timestamp_str,
                package_list_no,
                storage_service_id,
                fetch_job_id,
                run_as_task,
            )

    return processed_packages



from datetime import datetime

import pytest
import typesense
from pytz import timezone

from AIPscan import test_helpers
from AIPscan import typesense_helpers as ts_helpers
from AIPscan import typesense_test_helpers
from AIPscan.models import File
from AIPscan.models import Pipeline


def test_typesense_enabled(app_instance, enable_typesense):
    assert ts_helpers.typesense_enabled() is True


def test_client(app_instance, enable_typesense):
    client = ts_helpers.client()

    assert type(client) is typesense.Client
    assert client.config.nodes[0].host == "localhost"
    assert client.config.nodes[0].port == "8108"
    assert client.config.nodes[0].protocol == "http"
    assert client.config.api_key == "x0x0x0x0x0"


def test_search(app_instance, enable_typesense, mocker):
    client = ts_helpers.client()

    search_parameters = {"q": "*"}

    fake_results = {
        "facet_counts": [],
        "found": 1,
        "hits": [{"document": {"aip_id": 1}, "highlight": {}, "highlights": []}],
        "out_of": 252246,
        "page": 1,
        "request_params": {"collection_name": "aipscan_file", "per_page": 10, "q": "*"},
        "search_cutoff": False,
        "search_time_ms": 0,
    }

    fake_collection = typesense_test_helpers.FakeCollection(fake_results)

    query = mocker.patch("typesense.collections.Collections.__getitem__")
    query.return_value = fake_collection

    results = ts_helpers.search("file", search_parameters, client)

    assert results == fake_results


def test_collection_prefix(app_instance):
    fake_table_name = "test"

    prefixed = ts_helpers.collection_prefix(fake_table_name)

    assert prefixed == f"aipscan_{fake_table_name}"


@pytest.mark.parametrize(
    "datetime_obj,expected",
    [
        # PST midnight localized: exactly at the local midnight in US/Pacific.
        # The function should treat this as the PST date boundary and return the
        # corresponding epoch seconds
        (
            timezone("US/Pacific").localize(datetime(2019, 1, 30, 0, 0, 0)),
            1548835200,
        ),
        # PST early-morning time: a time after midnight in PST should be
        # truncated to the same PST date boundary.
        (
            timezone("US/Pacific").localize(datetime(2019, 1, 30, 2, 30, 50)),
            1548835200,
        ),
        # Naive datetime (no tz): for these tests naive datetimes are treated as
        # system/UTC-local time, so midnight naive here corresponds to the UTC
        # date boundary.
        (datetime(2019, 1, 30, 0, 0, 0), 1548806400),
        # Explicit UTC timezone: should map directly to the UTC date boundary
        # timestamp.
        (
            timezone("UTC").localize(datetime(2019, 1, 30, 0, 0, 0)),
            1548806400,
        ),
        # Naive later time in the day: later-naive times should be truncated to
        # that day's UTC date boundary.
        (datetime(2019, 1, 30, 2, 30, 50), 1548806400),
    ],
)
def test_datetime_to_timestamp_int(datetime_obj, expected):
    # Ensure datetime_to_timestamp_int converts a datetime to the start-of-day
    # (midnight in its timezone) before turning it into epoch seconds.
    assert ts_helpers.datetime_to_timestamp_int(datetime_obj) == expected


def test_assemble_filter_by():
    filters = [("storage_service_id", "=", 1), ("file_type", "=", "'original'")]
    filter_string = "storage_service_id:=1 && file_type:='original'"

    assert ts_helpers.assemble_filter_by(filters) == filter_string


def test_file_filters(app_instance):
    # Get local time zone then set time zone for testing
    local_timezone = test_helpers.set_timezone_and_return_current_timezone(
        "America/Vancouver"
    )

    file_filters = ts_helpers.file_filters(
        1, 1, datetime(2019, 1, 1), datetime(2019, 1, 31)
    )

    desired_filters = [
        ("aip_create_date", ">=", 1546300800),
        ("aip_create_date", "<", 1548892800),
        ("storage_service_id", "=", 1),
        ("file_type", "=", "'original'"),
        ("storage_location_id", "=", 1),
    ]

    assert file_filters == desired_filters

    # Reset time zone
    test_helpers.set_timezone(local_timezone)


def test_facet_value_counts():
    results = {
        "facet_counts": [
            {
                "field_name": "color",
                "counts": [{"value": "blue", "count": 9}, {"value": "red", "count": 7}],
            }
        ]
    }

    counts = ts_helpers.facet_value_counts(results)

    assert counts == {"color": {"blue": 9, "red": 7}}


def test_get_model_table():
    assert ts_helpers.get_model_table(File) == "file"


def test_collection_fields_from_model():
    pipeline_fields = [
        {"name": "id", "type": "int32"},
        {"name": "origin_pipeline", "type": "string", "optional": True},
        {"name": "dashboard_url", "type": "string", "optional": True},
    ]

    assert ts_helpers.collection_fields_from_model(Pipeline) == pipeline_fields


def test_initialize_index(app_instance, enable_typesense, mocker):
    class FakeCollections:
        def create(self):
            return True

    fake_collections = FakeCollections()
    fake_collection = typesense_test_helpers.FakeCollection()

    query = mocker.patch("typesense.collections.Collections.__getitem__")
    query.return_value = fake_collection

    query2 = mocker.patch("typesense.collections.Collections.create")
    query2.return_value = fake_collections

    ts_helpers.initialize_index()


def test_populate_index(app_instance, enable_typesense, mocker):
    fake_collection = typesense_test_helpers.FakeCollection()

    query = mocker.patch("typesense.collections.Collections.__getitem__")
    query.return_value = fake_collection

    ts_helpers.populate_index()


def test_augment_file_document_with_aip_data(app_instance):
    document = {}

    aip_cache = {
        "original_file_count": {3: 23},
        "storage_service_id": {3: 1},
        "storage_location_id": {3: [2]},
        "create_date": {3: datetime(2024, 3, 2, 5, 35)},
        "uuid": {3: "111111111111-1111-1111-11111111"},
        "transfer_name": {3: "Test AIP"},
    }

    class MockFileResult:
        id = 4
        aip_id = 3

    file_result = MockFileResult()

    # Get local time zone then set time zone for testing
    local_timezone = test_helpers.set_timezone_and_return_current_timezone(
        "America/Vancouver"
    )

    ts_helpers.augment_file_document_with_aip_data(
        "file", document, aip_cache, file_result
    )

    assert document["storage_service_id"] == 1
    assert document["storage_location_id"] == [2]
    assert document["aip_create_date"] == 1709366400
    assert document["aip_uuid"] == "111111111111-1111-1111-11111111"
    assert document["transfer_name"] == "Test AIP"

    # Reset time zone
    test_helpers.set_timezone(local_timezone)



from flask import render_template
from flask import request

from AIPscan import typesense_helpers as ts_helpers
from AIPscan.Data import fields
from AIPscan.Data import report_data
from AIPscan.Data import report_data_typesense
from AIPscan.helpers import parse_bool
from AIPscan.helpers import parse_datetime_bound
from AIPscan.Reporter import download_csv
from AIPscan.Reporter import format_size_for_csv
from AIPscan.Reporter import get_display_end_date
from AIPscan.Reporter import reporter
from AIPscan.Reporter import request_params
from AIPscan.Reporter import translate_headers

HEADERS = [
    fields.FIELD_PUID,
    fields.FIELD_FORMAT,
    fields.FIELD_VERSION,
    fields.FIELD_COUNT,
    fields.FIELD_SIZE,
]


@reporter.route("/report_format_versions_count/", methods=["GET"])
def report_format_versions_count():
    """Return overview of format versions in Storage Service."""
    storage_service_id = request.args.get(request_params.STORAGE_SERVICE_ID)
    storage_location_id = request.args.get(request_params.STORAGE_LOCATION_ID)
    start_date = parse_datetime_bound(request.args.get(request_params.START_DATE))
    end_date = parse_datetime_bound(
        request.args.get(request_params.END_DATE), upper=True
    )
    csv = parse_bool(request.args.get(request_params.CSV), default=False)

    if ts_helpers.typesense_enabled():
        version_data = report_data_typesense.format_versions_count(
            storage_service_id=storage_service_id,
            start_date=start_date,
            end_date=end_date,
            storage_location_id=storage_location_id,
        )
    else:
        version_data = report_data.format_versions_count(
            storage_service_id=storage_service_id,
            start_date=start_date,
            end_date=end_date,
            storage_location_id=storage_location_id,
        )

    versions = version_data.get(fields.FIELD_FORMAT_VERSIONS)

    if csv:
        headers = translate_headers(HEADERS, True)

        filename = "format_versions.csv"
        csv_data = format_size_for_csv(versions)
        return download_csv(headers, csv_data, filename)

    headers = translate_headers(HEADERS)

    return render_template(
        "report_format_versions_count.html",
        storage_service_id=storage_service_id,
        storage_service_name=version_data.get(fields.FIELD_STORAGE_NAME),
        storage_location_description=version_data.get(fields.FIELD_STORAGE_LOCATION),
        columns=translate_headers(headers),
        versions=versions,
        total_file_count=sum(
            version.get(fields.FIELD_COUNT, 0) for version in versions
        ),
        total_size=sum(version.get(fields.FIELD_SIZE, 0) for version in versions),
        puid_count=len(versions),
        start_date=start_date,
        end_date=get_display_end_date(end_date),
    )



"""Data namespace

Return 'unkempt' data from AIPscan so that it can be filtered, cut, and
remixed as the caller desires. Data is 'unkempt' not raw. No data is
raw. No data is without bias.
"""

from flask import request
from flask_restx import Namespace
from flask_restx import Resource

from AIPscan.API import fields
from AIPscan.Data import data
from AIPscan.helpers import parse_bool
from AIPscan.helpers import parse_datetime_bound

api = Namespace("data", description="Retrieve data from AIPscan to shape as you desire")

"""
data = api.model('Data', {
    'id': fields.String(required=True, description='Do we need this? An identifier for the data...'),
    'name': fields.String(required=True, description='Do we need this? A name for the datas...'),
})
"""


@api.route("/storage-services")
class StorageServices(Resource):
    def get(self):
        """Return data on storage services."""
        return data.storage_services()


@api.route("/aip-overview/<storage_service_id>")
class AIPList(Resource):
    @api.doc(
        "list_formats",
        params={
            fields.FIELD_ORIGINAL_FILES: {
                "description": "Return data on original files or preservation derivatives",
                "in": "query",
                "type": "bool",
            },
            fields.FIELD_STORAGE_LOCATION: {
                "description": "Storage Location ID",
                "in": "query",
                "type": "int",
            },
            fields.FIELD_START_DATE: {
                "description": "AIP creation start date (inclusive, YYYY-MM-DD)",
                "in": "query",
                "type": "str",
            },
            fields.FIELD_END_DATE: {
                "description": "AIP creation end date (inclusive, YYYY-MM-DD)",
                "in": "query",
                "type": "str",
            },
        },
    )
    def get(self, storage_service_id):
        """Return data on AIPs and the file formats they contain."""
        original_files = parse_bool(request.args.get(fields.FIELD_ORIGINAL_FILES, True))
        storage_location_id = request.args.get(fields.FIELD_STORAGE_LOCATION)
        start_date = parse_datetime_bound(request.args.get(fields.FIELD_START_DATE))
        end_date = parse_datetime_bound(
            request.args.get(fields.FIELD_END_DATE), upper=True
        )
        return data.aip_file_format_overview(
            storage_service_id=storage_service_id,
            start_date=start_date,
            end_date=end_date,
            original_files=original_files,
            storage_location_id=storage_location_id,
        )


@api.route("/fmt-overview/<storage_service_id>")
class FMTList(Resource):
    @api.doc(
        "list_formats",
        params={
            fields.FIELD_ORIGINAL_FILES: {
                "description": "Return data on original files or preservation derivatives",
                "in": "query",
                "type": "bool",
            },
            fields.FIELD_STORAGE_LOCATION: {
                "description": "Storage Location ID",
                "in": "query",
                "type": "int",
            },
        },
    )
    def get(self, storage_service_id):
        """Return data on PUIDs and the AIPs they are contained within."""
        original_files = parse_bool(request.args.get(fields.FIELD_ORIGINAL_FILES, True))
        storage_location_id = request.args.get(fields.FIELD_STORAGE_LOCATION)
        return data.file_format_aip_overview(
            storage_service_id=storage_service_id,
            original_files=original_files,
            storage_location_id=storage_location_id,
        )


@api.route("/derivative-overview/<storage_service_id>")
class DerivativeList(Resource):
    @api.doc(
        "list_aips",
        params={
            fields.FIELD_STORAGE_LOCATION: {
                "description": "Storage Location ID",
                "in": "query",
                "type": "int",
            }
        },
    )
    def get(self, storage_service_id):
        """List original and derivative identifiers per AIP"""
        storage_location_id = request.args.get(fields.FIELD_STORAGE_LOCATION)
        return data.derivative_overview(
            storage_service_id=storage_service_id,
            storage_location_id=storage_location_id,
        )



import pytest

from AIPscan.conftest import ORIGINAL_FILE_SIZE
from AIPscan.conftest import PRESERVATION_FILE_SIZE
from AIPscan.conftest import TIFF_PUID
from AIPscan.Data import fields
from AIPscan.Data import report_data
from AIPscan.Data.tests import (
    MOCK_AIPS_BY_FORMAT_OR_PUID_QUERY_RESULTS as MOCK_QUERY_RESULTS,
)
from AIPscan.Data.tests import MOCK_STORAGE_SERVICE
from AIPscan.Data.tests import MOCK_STORAGE_SERVICE_ID
from AIPscan.Data.tests import MOCK_STORAGE_SERVICE_NAME
from AIPscan.test_helpers import create_test_storage_location


@pytest.mark.parametrize(
    "query_results, results_count",
    [
        # Empty result set, count is 0.
        ([], 0),
        # Test the return of complete result set, count is the length
        # of all results.
        (MOCK_QUERY_RESULTS, len(MOCK_QUERY_RESULTS)),
        # Test the return of only the first two results, count is 2.
        (MOCK_QUERY_RESULTS[:2], 2),
    ],
)
def test_aips_by_puid(app_instance, mocker, query_results, results_count):
    """Test that results match high-level expectations."""
    query = mocker.patch("AIPscan.Data.report_data._query_aips_by_file_format_or_puid")
    query.return_value = query_results

    get_ss = mocker.patch("AIPscan.Data._get_storage_service")
    get_ss.return_value = MOCK_STORAGE_SERVICE

    test_location = create_test_storage_location()
    get_location = mocker.patch("AIPscan.Data._get_storage_location")
    get_location.return_value = test_location

    report = report_data.aips_by_puid(
        storage_service_id=MOCK_STORAGE_SERVICE_ID,
        storage_location_id=test_location.id,
        puid="fmt/###",
    )
    assert report[fields.FIELD_STORAGE_NAME] == MOCK_STORAGE_SERVICE_NAME
    assert report[fields.FIELD_STORAGE_LOCATION] == test_location.description
    assert len(report[fields.FIELD_AIPS]) == results_count


@pytest.mark.parametrize(
    "test_aip", [mock_result for mock_result in MOCK_QUERY_RESULTS]
)
def test_aips_by_puid_aip_elements(app_instance, mocker, test_aip):
    """Test that structure of AIP data matches expectations."""
    mock_query = mocker.patch(
        "AIPscan.Data.report_data._query_aips_by_file_format_or_puid"
    )
    mock_query.return_value = [test_aip]

    mock_get_ss_name = mocker.patch("AIPscan.Data._get_storage_service")
    mock_get_ss_name.return_value = MOCK_STORAGE_SERVICE

    report = report_data.aips_by_puid(MOCK_STORAGE_SERVICE_ID, "fmt/###")
    report_aip = report[fields.FIELD_AIPS][0]

    assert test_aip.id == report_aip.get(fields.FIELD_ID)
    assert test_aip.name == report_aip.get(fields.FIELD_AIP_NAME)
    assert test_aip.uuid == report_aip.get(fields.FIELD_UUID)
    assert test_aip.file_count == report_aip.get(fields.FIELD_COUNT)
    assert test_aip.total_size == report_aip.get(fields.FIELD_SIZE)


@pytest.mark.parametrize(
    "puid, original_files, aip_count, total_file_count, total_file_size",
    [
        # Original format in test data should be included in results.
        (TIFF_PUID, True, 1, 1, ORIGINAL_FILE_SIZE),
        # Original format not in test data should not be included in
        # results.
        ("fmt/1", True, 0, 0, 0),
        # Preservation format in test data should be included in
        # results.
        (TIFF_PUID, False, 1, 1, PRESERVATION_FILE_SIZE),
        # Non-existent preservation format not present in test data
        # should not be included in results.
        ("x-fmt/999999", False, 0, 0, 0),
        # Passing a None value for original_files returns original
        # files.
        (TIFF_PUID, None, 1, 1, ORIGINAL_FILE_SIZE),
        # Passing an incorrect value for original_files returns
        # original files.
        (TIFF_PUID, "invalid", 1, 1, ORIGINAL_FILE_SIZE),
    ],
)
def test_aips_by_file_format_contents(
    app_with_populated_files,
    puid,
    original_files,
    aip_count,
    total_file_count,
    total_file_size,
):
    """Test that content of response matches expectations.

    This integration test uses a pre-populated fixture to verify that
    the database access layer of our endpoint returns what we expect.
    """
    results = report_data.aips_by_puid(
        storage_service_id=1, puid=puid, original_files=original_files
    )
    aips = results[fields.FIELD_AIPS]
    assert len(aips) == aip_count
    assert sum(aip[fields.FIELD_COUNT] for aip in aips) == total_file_count
    assert sum(aip[fields.FIELD_SIZE] for aip in aips) == total_file_size



"""Tests for METS helper functions."""

import os

import metsrw
import pytest

from AIPscan.Aggregator import mets_parse_helpers

FIXTURES_DIR = "fixtures"


@pytest.mark.parametrize(
    "fixture_path, transfer_name, mets_error",
    [
        (os.path.join("features_mets", "features-mets.xml"), "myTransfer", False),
        (os.path.join("iso_mets", "iso_mets.xml"), "iso", False),
        (
            os.path.join("original_name_mets", "document-empty-dirs.xml"),
            "empty-dirs",
            False,
        ),
        # Exception: Cannot disambiguate dmdSec_1 in the METS using
        # mets-reader-writer and so we cannot retrieve the originalName.
        (os.path.join("original_name_mets", "dataverse_example.xml"), "", True),
    ],
)
def test_get_aip_original_name(fixture_path, transfer_name, mets_error):
    """Make sure that we can reliably get original name from the METS
    file given we haven't any mets-reader-writer helpers.
    """
    script_dir = os.path.dirname(os.path.realpath(__file__))
    mets_file = os.path.join(script_dir, FIXTURES_DIR, fixture_path)
    mets = metsrw.METSDocument.fromfile(mets_file)
    if mets_error:
        # Function should raise an error to work with.
        with pytest.raises(mets_parse_helpers.METSError):
            _ = mets_parse_helpers.get_aip_original_name(mets)
        return
    assert mets_parse_helpers.get_aip_original_name(mets) == transfer_name
    # Test the same works with a string.
    with open(mets_file, "rb") as mets_stream:
        mets = metsrw.METSDocument.fromstring(mets_stream.read())
        assert mets_parse_helpers.get_aip_original_name(mets) == transfer_name
    # Use that string to manipulate the text so that the element cannot
    # be found.
    with open(mets_file, "rb") as mets_stream:
        # Remove originalName and test for an appropriate side-effect.
        new_mets = mets_stream.read()
        new_mets = new_mets.replace(b"originalName", b"originalNameIsNotHere")
        mets = metsrw.METSDocument.fromstring(new_mets)
        # Function should raise an error to work with.
        with pytest.raises(mets_parse_helpers.METSError):
            _ = mets_parse_helpers.get_aip_original_name(mets) == transfer_name



import logging
from functools import wraps

from AIPscan import db
from AIPscan.models import package_tasks

logger = logging.getLogger(__name__)


def write_celery_update(package_lists_task, workflow_coordinator):
    # Write status update to celery model.
    package_task = package_tasks(
        package_task_id=package_lists_task.id,
        workflow_coordinator_id=workflow_coordinator.request.id,
    )
    db.session.add(package_task)
    db.session.commit()


def with_db_session(func):
    """Ensure the scoped session is cleaned up around each call."""

    @wraps(func)
    def wrapper(*args, **kwargs):
        func_name = func.__name__
        try:
            return func(*args, **kwargs)
        except Exception:
            logger.exception("Unhandled exception in %s", func_name)
            try:
                db.session.rollback()
                logger.debug("Rolled back session after error in %s", func_name)
            except Exception:
                logger.exception(
                    "Failed to rollback session after error in %s", func_name
                )
            raise
        finally:
            try:
                db.session.remove()
            except Exception:
                logger.exception(
                    "Failed to dispose session after call to %s", func_name
                )

    return wrapper



from AIPscan import db
from AIPscan.Data import fields
from AIPscan.models import StorageLocation
from AIPscan.models import StorageService


def _get_storage_service(storage_service_id):
    """Return Storage Service with ID or None.

    Unlike elsewhere in our application, here we do not fall back to
    a different StorageService if the user-supplied ID is invalid to
    prevent inaccurate information from being returned.

    :param storage_service_id: Storage Service ID

    :returns: StorageService object or None
    """
    return db.session.get(StorageService, storage_service_id)


def get_storage_service_name(storage_service_id):
    """Return name of Storage Service or None."""
    name = None
    if storage_service_id:
        storage_service = _get_storage_service(storage_service_id)
        if storage_service:
            name = storage_service.name
    return name


def _get_storage_location(storage_location_id):
    """Return Storage Location with ID or None."""
    return db.session.get(StorageLocation, storage_location_id)


def get_storage_location_description(storage_location_id):
    """Return description of Storage Location or None."""
    description = None
    if storage_location_id:
        storage_location = _get_storage_location(storage_location_id)
        if storage_location:
            description = storage_location.description
    return description


def report_dict(storage_service_id, storage_location_id):
    report = {}

    report[fields.FIELD_STORAGE_NAME] = get_storage_service_name(storage_service_id)
    report[fields.FIELD_STORAGE_LOCATION] = get_storage_location_description(
        storage_location_id
    )

    return report



import csv
import datetime
import uuid

import pytest

from AIPscan.Data import fields
from AIPscan.Data.report_data import aips_by_file_format
from AIPscan.Data.tests import (
    MOCK_AIPS_BY_FORMAT_OR_PUID_QUERY_RESULTS as QUERY_RESULTS,
)
from AIPscan.Data.tests import MOCK_STORAGE_SERVICE as STORAGE_SERVICE
from AIPscan.Data.tests import MOCK_STORAGE_SERVICE_ID as STORAGE_SERVICE_ID
from AIPscan.models import File
from AIPscan.models import FileType
from AIPscan.Reporter import helpers
from AIPscan.Reporter.report_aip_contents import CSV_HEADERS as AIP_CONTENTS_HEADERS
from AIPscan.Reporter.report_aips_by_format import HEADERS as AIPS_BY_FORMAT_HEADERS

ROWS_WITH_SIZE = [
    {
        fields.FIELD_AIP_UUID: "test uuid",
        fields.FIELD_AIP_NAME: "test name",
        fields.FIELD_SIZE: 1560321,
    },
    {
        fields.FIELD_AIP_UUID: "test uuid2",
        fields.FIELD_AIP_NAME: "test name2",
        fields.FIELD_SIZE: 123423,
    },
]

ROWS_WITH_SIZE_FORMATTED = [
    {
        fields.FIELD_AIP_UUID: "test uuid",
        fields.FIELD_AIP_NAME: "test name",
        fields.FIELD_SIZE: "1.6 MB",
        fields.FIELD_SIZE_BYTES: 1560321,
    },
    {
        fields.FIELD_AIP_UUID: "test uuid2",
        fields.FIELD_AIP_NAME: "test name2",
        fields.FIELD_SIZE: "123.4 kB",
        fields.FIELD_SIZE_BYTES: 123423,
    },
]

ROWS_WITHOUT_SIZE = [
    {fields.FIELD_NAME: "test name", fields.FIELD_UUID: "test uuid"},
    {fields.FIELD_NAME: "test name2", fields.FIELD_UUID: "test uuid2"},
]


def test_download_csv(app_instance, mocker):
    """Test downloading report data as CSV."""
    CSV_FILE = "test.csv"
    CONTENT_DISPOSITION = f"attachment; filename={CSV_FILE}"
    CSV_MIMETYPE = "text/csv"

    query_results = QUERY_RESULTS[:2]

    mock_query = mocker.patch(
        "AIPscan.Data.report_data._query_aips_by_file_format_or_puid"
    )
    mock_query.return_value = query_results
    mock_get_ss_name = mocker.patch("AIPscan.Data._get_storage_service")
    mock_get_ss_name.return_value = STORAGE_SERVICE

    headers = helpers.translate_headers(AIPS_BY_FORMAT_HEADERS)

    report_data = aips_by_file_format(STORAGE_SERVICE_ID, "test")
    response = helpers.download_csv(headers, report_data[fields.FIELD_AIPS], CSV_FILE)

    # Assert response is shaped as we expect.
    assert response is not None
    assert response.headers["Content-Disposition"] == CONTENT_DISPOSITION
    assert response.mimetype == CSV_MIMETYPE

    # Assert that content of CSV file is what we expect.
    content = response.get_data().decode()
    reader = csv.reader(content.splitlines())
    line_count = 0
    for row in reader:
        if line_count == 0:
            assert row == headers
        elif line_count == 1:
            assert row[0] == "aip0"
            assert row[1] == "11111111-1111-1111-1111-111111111111"
            assert row[2] == "5"
            assert row[3] == "12345678"
        else:
            assert row[0] == "aip1"
            assert row[1] == "22222222-2222-2222-2222-222222222222"
            assert row[2] == "3"
            assert row[3] == "123456"
        line_count += 1
    assert line_count == len(query_results) + 1


@pytest.mark.parametrize(
    "data,expected_output",
    [
        # No adding of header for size in bytes
        (
            {"headers": AIPS_BY_FORMAT_HEADERS, "add_bytes_column": False},
            ["AIP Name", "UUID", "Count", "Size"],
        ),
        # Adding of header for size in bytes at end of header list
        (
            {"headers": AIPS_BY_FORMAT_HEADERS, "add_bytes_column": True},
            ["AIP Name", "UUID", "Count", "Size", "Size (bytes)"],
        ),
        # Adding of header for size in bytes not at end of header list
        (
            {"headers": AIP_CONTENTS_HEADERS, "add_bytes_column": True},
            ["UUID", "AIP Name", "Created Date", "Size", "Size (bytes)", "Formats"],
        ),
    ],
)
def test_translate_headers(data, expected_output):
    headers = helpers.translate_headers(data["headers"], data["add_bytes_column"])

    assert headers == expected_output


@pytest.mark.parametrize(
    "data,expected_output",
    [
        # Dicts with size key should have the corresponding value
        # formatted by filesizeformat.
        (ROWS_WITH_SIZE, ROWS_WITH_SIZE_FORMATTED),
        # Dicts without size key should be unaffected.
        (ROWS_WITHOUT_SIZE, ROWS_WITHOUT_SIZE),
    ],
)
def test_format_size_for_csv(data, expected_output):
    """Test formatting size in report endpoint data."""
    assert helpers.format_size_for_csv(data) == expected_output


def test_get_premis_xml_lines():
    premis_xml = "First line\nSecond line"

    file_ = File(
        uuid=uuid.uuid4(),
        name="test.txt",
        size=12345,
        aip_id=2,
        file_type=FileType.original,
        file_format="Plain Text File",
        puid="x-fmt/111",
        filepath="/path/to/file.txt",
        date_created=datetime.datetime.now(),
        checksum_type="md5",
        checksum_value="anotherfakemd5",
        premis_object=premis_xml,
    )

    lines = file_.get_premis_xml_lines()

    assert len(lines) == 2


@pytest.mark.parametrize(
    "paging",
    [
        # Test paging window at start of results
        {"per_page": 5, "total": 17, "page": 1, "first_item": 1, "last_item": 5},
        # Test paging window on an arbitrary page of results
        {"per_page": 5, "total": 17, "page": 3, "first_item": 11, "last_item": 15},
        # Test paging window with last item being set to the total of items
        {"per_page": 5, "total": 17, "page": 4, "first_item": 16, "last_item": 17},
        # Test paging window at start of results with incomplete page
        {"per_page": 7, "total": 4, "page": 1, "first_item": 1, "last_item": 4},
    ],
)
def test_calculate_paging_window(paging):
    class MockPagination:
        pass

    pagination = MockPagination()
    pagination.page = paging["page"]
    pagination.per_page = paging["per_page"]
    pagination.total = paging["total"]

    first_item, last_item = helpers.calculate_paging_window(pagination)

    assert first_item == paging["first_item"]
    assert last_item == paging["last_item"]


def test_remove_dict_none_values():
    test_values = {"foo": "bar", "nada": None}

    querystring = helpers.remove_dict_none_values(test_values)

    assert querystring == {"foo": "bar", "nada": ""}



from AIPscan import typesense_helpers as ts_helpers
from AIPscan.Data import fields
from AIPscan.Data import report_dict


def document_to_report_row(document, report_fields):
    row = {}

    for report_field in report_fields:
        document_field = report_fields[report_field]

        if document_field == "size":
            row[report_field] = int(document.get(document_field, 0))
        else:
            row[report_field] = document.get(document_field, "")

    return row


def formats_count(
    storage_service_id,
    storage_location_id,
    start_date,
    end_date,
    include_size_data=True,
):
    report = report_dict(storage_service_id, storage_location_id)
    report[fields.FIELD_FORMATS] = []

    # Get format counts via facet data
    file_filters = ts_helpers.file_filters(
        storage_service_id, storage_location_id, start_date, end_date
    )

    results = ts_helpers.search(
        "file",
        {
            "q": "*",
            "filter_by": ts_helpers.assemble_filter_by(file_filters),
            "exclude_fields": "*",
            "facet_by": "file_format",
            "max_facet_values": 10000,
        },
    )

    value_counts = ts_helpers.facet_value_counts(results, "file_format")

    # If no formats were found then don't proceed to get counts
    if len(value_counts) == 0:
        return report

    format_size_sums = {}
    if include_size_data:
        # Request total size of files for each file format
        search_requests = {"searches": []}
        for file_format in value_counts.keys():
            format_filters = file_filters.copy()
            format_filters.append(("file_format", "=", f"`{file_format}`"))

            format_filter_by = ts_helpers.assemble_filter_by(format_filters)

            search_requests["searches"].append(
                {
                    "collection": ts_helpers.collection_prefix("file"),
                    "q": "*",
                    "include_fields": "file_format",
                    "filter_by": format_filter_by,
                    "facet_by": "size",
                    "max_facet_values": 1,
                }
            )

        searches = ts_helpers.client().multi_search.perform(
            search_requests, {"limit_multi_searches": len(search_requests["searches"])}
        )

        # Summarize file format sizes
        for results in searches["results"]:
            if "hits" in results:
                file_format = results["hits"][0]["document"]["file_format"]

                for count in results["facet_counts"]:
                    if count["field_name"] == "size":
                        format_size_sums[file_format] = count["stats"]["sum"]

    # Format results for report
    format_data = {}

    for file_format in value_counts.keys():
        format_data[file_format] = {
            fields.FIELD_FORMAT: file_format,
            fields.FIELD_COUNT: value_counts[file_format],
        }

        if format_size_sums != {}:
            format_data[file_format][fields.FIELD_SIZE] = format_size_sums.get(
                file_format, 0
            )

    report[fields.FIELD_FORMATS] = list(format_data.values())

    return report


def largest_aips(
    storage_service_id, start_date, end_date, storage_location_id, limit=20
):
    report = report_dict(storage_service_id, storage_location_id)
    report[fields.FIELD_AIPS] = []

    # Assemble filter_by
    start_timestamp = ts_helpers.datetime_to_timestamp_int(start_date)
    end_timestamp = ts_helpers.datetime_to_timestamp_int(end_date)

    filters = [
        ("create_date", ">=", start_timestamp),
        ("create_date", "<", end_timestamp),
        ("storage_service_id", "=", storage_service_id),
    ]

    if storage_location_id is not None and storage_location_id != "":
        filters.append(("storage_location_id", "=", storage_location_id))

    filter_by = ts_helpers.assemble_filter_by(filters)

    # Get AIPs in descending order of size
    report_fields = {
        fields.FIELD_UUID: "uuid",
        fields.FIELD_NAME: "transfer_name",
        fields.FIELD_SIZE: "size",
        fields.FIELD_FILE_COUNT: "original_file_count",
    }

    results = ts_helpers.search(
        "aip",
        {
            "q": "*",
            "filter_by": filter_by,
            "include_fields": ",".join(list(report_fields.values())),
            "per_page": limit,
            "sort_by": "size:desc",
        },
    )

    # Format results for report
    for hit in results["hits"]:
        file_info = document_to_report_row(hit["document"], report_fields)
        report[fields.FIELD_AIPS].append(file_info)

    return report


def largest_files(
    storage_service_id,
    start_date,
    end_date,
    storage_location_id,
    file_type=None,
    limit=20,
):
    report = report_dict(storage_service_id, storage_location_id)
    report[fields.FIELD_FILES] = []

    # Get files in descending order of size
    file_filters = ts_helpers.file_filters(
        storage_service_id, storage_location_id, start_date, end_date
    )

    report_fields = {
        fields.FIELD_ID: "id",
        fields.FIELD_UUID: "uuid",
        fields.FIELD_NAME: "name",
        fields.FIELD_SIZE: "size",
        fields.FIELD_FILE_TYPE: "file_type",
        fields.FIELD_FORMAT: "file_format",
        fields.FIELD_VERSION: "format_version",
        fields.FIELD_PUID: "puid",
        fields.FIELD_AIP_NAME: "transfer_name",
        fields.FIELD_AIP_UUID: "aip_uuid",
    }

    results = ts_helpers.search(
        "file",
        {
            "q": "*",
            "filter_by": ts_helpers.assemble_filter_by(file_filters),
            "include_fields": ",".join(list(report_fields.values())),
            "per_page": limit,
            "sort_by": "size:desc",
        },
    )

    # Format results for report
    for hit in results["hits"]:
        file_info = document_to_report_row(hit["document"], report_fields)
        report[fields.FIELD_FILES].append(file_info)

    return report


def format_versions_count(
    storage_service_id, start_date, end_date, storage_location_id=None
):
    report = report_dict(storage_service_id, storage_location_id)
    report[fields.FIELD_FORMAT_VERSIONS] = []

    # Get format counts via facet data
    file_filters = ts_helpers.file_filters(
        storage_service_id, storage_location_id, start_date, end_date
    )

    results = ts_helpers.search(
        "file",
        {
            "q": "*",
            "filter_by": ts_helpers.assemble_filter_by(file_filters),
            "include_fields": "puid",
            "facet_by": "puid",
            "max_facet_values": 10000,
        },
    )

    puid_counts = ts_helpers.facet_value_counts(results, "puid")

    # If no format versions were found then don't proceed to get counts
    if len(puid_counts) == 0:
        return report

    # Request total size of files for each PUID
    puids = list(puid_counts.keys())

    search_requests = {"searches": []}
    for puid in puids:
        format_filters = file_filters.copy()
        format_filters.append(("puid", "=", f"`{puid}`"))

        format_filter_by = ts_helpers.assemble_filter_by(format_filters)

        search_requests["searches"].append(
            {
                "collection": ts_helpers.collection_prefix("file"),
                "q": "*",
                "include_fields": "puid,file_format,format_version",
                "filter_by": format_filter_by,
                "facet_by": "size",
                "max_facet_values": 1,
            }
        )

    searches = ts_helpers.client().multi_search.perform(
        search_requests, {"limit_multi_searches": len(search_requests["searches"])}
    )

    # Summarize PUID sizes
    puid_sizes = {}
    puid_formats = {}
    puid_versions = {}

    for results in searches["results"]:
        document = results["hits"][0]["document"]

        puid = document["puid"]
        puid_formats[puid] = document["file_format"]
        puid_versions[puid] = document.get("format_version", "")

        for count in results["facet_counts"]:
            if count["field_name"] == "size":
                puid_sizes[puid] = count["stats"]["sum"]

    # Format results for report
    for puid in puids:
        version_info = {}

        version_info[fields.FIELD_PUID] = puid
        version_info[fields.FIELD_FORMAT] = puid_formats[puid]
        version_info[fields.FIELD_VERSION] = puid_versions.get("puid", "")
        version_info[fields.FIELD_COUNT] = puid_counts[puid]

        version_info[fields.FIELD_SIZE] = 0

        if puid_sizes[puid] is not None:
            version_info[fields.FIELD_SIZE] = puid_sizes[puid]

        report[fields.FIELD_FORMAT_VERSIONS].append(version_info)

    return report



from AIPscan import typesense_test_helpers

FAKE_RESULTS = {"found": 1}


def test_fake_documents():
    fake_documents = typesense_test_helpers.FakeDocuments(FAKE_RESULTS)

    assert fake_documents.fake_results == FAKE_RESULTS
    assert fake_documents.search({}) == FAKE_RESULTS


def test_fake_collection():
    fake_collection = typesense_test_helpers.FakeCollection(FAKE_RESULTS)

    assert fake_collection.documents.fake_results == FAKE_RESULTS
    assert fake_collection.documents.search({}) == FAKE_RESULTS



from datetime import datetime

import pytest

from AIPscan.conftest import STORAGE_LOCATION_1_DESCRIPTION
from AIPscan.conftest import STORAGE_LOCATION_1_UUID
from AIPscan.conftest import STORAGE_LOCATION_2_DESCRIPTION
from AIPscan.conftest import STORAGE_LOCATION_2_UUID
from AIPscan.Data import fields
from AIPscan.Data import report_data
from AIPscan.helpers import parse_datetime_bound

ORDERED_LIST = [{fields.FIELD_AIPS: 8}, {fields.FIELD_AIPS: 3}]
UNORDERED_LIST = [{fields.FIELD_AIPS: 3}, {fields.FIELD_AIPS: 8}]

DATE_BEFORE_AIP_1 = "2019-01-01"
DATE_AFTER_AIP_1 = "2020-01-02"
DATE_BEFORE_AIP_2 = "2020-05-30"
DATE_AFTER_AIP_2 = "2020-06-02"
DATE_BEFORE_AIP_3 = "2021-05-30"
DATE_AFTER_AIP_3 = "2021-06-01"


@pytest.mark.parametrize(
    "input_list, expected_output",
    [
        # Test that out-of-order list is sorted.
        (UNORDERED_LIST, ORDERED_LIST),
        # Test that list already in desired order is kept as-is.
        (ORDERED_LIST, ORDERED_LIST),
        # Test empty list.
        ([], []),
    ],
)
def test_sort_storage_locations(input_list, expected_output):
    assert report_data._sort_storage_locations(input_list) == expected_output


@pytest.mark.parametrize(
    "storage_service_id, storage_service_name, start_date, end_date, locations_count, location_1_aips_count, location_1_total_size, location_1_file_count, location_2_aips_count, location_2_total_size, location_2_file_count",
    [
        # Request for a Storage Service populated with two locations, each
        # containing AIPs and files.
        (
            1,
            "test storage service",
            DATE_BEFORE_AIP_1,
            DATE_AFTER_AIP_3,
            2,
            2,
            1600,
            3,
            1,
            5000,
            2,
        ),
        # Request for a non-existent Storage Service.
        (4, None, DATE_BEFORE_AIP_1, DATE_BEFORE_AIP_3, 0, 0, 0, 0, 0, 0, 0),
        # Request for a None Storage Service.
        (None, None, DATE_BEFORE_AIP_1, DATE_BEFORE_AIP_3, 0, 0, 0, 0, 0, 0, 0),
        # Test date filtering.
        (
            1,
            "test storage service",
            DATE_AFTER_AIP_1,
            DATE_AFTER_AIP_2,
            2,
            1,
            1000,
            1,
            0,
            0,
            0,
        ),
        (
            1,
            "test storage service",
            DATE_BEFORE_AIP_1,
            DATE_BEFORE_AIP_2,
            2,
            1,
            600,
            2,
            0,
            0,
            0,
        ),
        (
            1,
            "test storage service",
            DATE_BEFORE_AIP_1,
            DATE_BEFORE_AIP_3,
            2,
            2,
            1600,
            3,
            0,
            0,
            0,
        ),
        (
            1,
            "test storage service",
            DATE_AFTER_AIP_2,
            DATE_AFTER_AIP_3,
            2,
            0,
            0,
            0,
            1,
            5000,
            2,
        ),
    ],
)
def test_storage_locations_data(
    storage_locations,
    storage_service_id,
    storage_service_name,
    start_date,
    end_date,
    locations_count,
    location_1_aips_count,
    location_1_total_size,
    location_1_file_count,
    location_2_aips_count,
    location_2_total_size,
    location_2_file_count,
):
    """Test response from report_data.storage_locations endpoint."""
    report = report_data.storage_locations(
        storage_service_id=storage_service_id, start_date=start_date, end_date=end_date
    )

    assert report[fields.FIELD_STORAGE_NAME] == storage_service_name

    locations = report[fields.FIELD_LOCATIONS]
    assert len(locations) == locations_count

    if locations_count < 1:
        return

    first_location = locations[0]
    # Account for sorting changes made possible by date filtering.
    if location_1_aips_count > location_2_aips_count:
        assert first_location[fields.FIELD_UUID] == STORAGE_LOCATION_1_UUID
        assert (
            first_location[fields.FIELD_STORAGE_LOCATION]
            == STORAGE_LOCATION_1_DESCRIPTION
        )
        assert first_location[fields.FIELD_AIPS] == location_1_aips_count
        assert first_location[fields.FIELD_SIZE] == location_1_total_size
        assert first_location[fields.FIELD_FILE_COUNT] == location_1_file_count
    else:
        assert first_location[fields.FIELD_UUID] == STORAGE_LOCATION_2_UUID
        assert (
            first_location[fields.FIELD_STORAGE_LOCATION]
            == STORAGE_LOCATION_2_DESCRIPTION
        )
        assert first_location[fields.FIELD_AIPS] == location_2_aips_count
        assert first_location[fields.FIELD_SIZE] == location_2_total_size
        assert first_location[fields.FIELD_FILE_COUNT] == location_2_file_count

    second_location = locations[1]
    # Account for sorting changes made possible by date filtering.
    if location_1_aips_count > location_2_aips_count:
        assert second_location[fields.FIELD_UUID] == STORAGE_LOCATION_2_UUID
        assert (
            second_location[fields.FIELD_STORAGE_LOCATION]
            == STORAGE_LOCATION_2_DESCRIPTION
        )
        assert second_location[fields.FIELD_AIPS] == location_2_aips_count
        assert second_location[fields.FIELD_SIZE] == location_2_total_size
        assert second_location[fields.FIELD_FILE_COUNT] == location_2_file_count
    else:
        assert second_location[fields.FIELD_UUID] == STORAGE_LOCATION_1_UUID
        assert (
            second_location[fields.FIELD_STORAGE_LOCATION]
            == STORAGE_LOCATION_1_DESCRIPTION
        )
        assert second_location[fields.FIELD_AIPS] == location_1_aips_count
        assert second_location[fields.FIELD_SIZE] == location_1_total_size
        assert second_location[fields.FIELD_FILE_COUNT] == location_1_file_count


@pytest.mark.parametrize(
    "start_date, end_date, expected_output",
    [
        # Test span within month.
        (
            "2022-01-01",
            "2022-01-04",
            ["2022-01-01", "2022-01-02", "2022-01-03", "2022-01-04"],
        ),
        # Test span across months.
        (
            "2022-02-27",
            "2022-03-02",
            ["2022-02-27", "2022-02-28", "2022-03-01", "2022-03-02"],
        ),
        # Test what happens if end is before start.
        ("2022-06-01", "2022-05-01", []),
        # Test that datetime.min is replaced by earliest AIP created date
        (datetime.min, "2020-01-03", ["2020-01-01", "2020-01-02", "2020-01-03"]),
        # Test that datetime.max is replaced by datetime.now()
        (
            datetime.now().isoformat()[:10],
            datetime.max,
            [datetime.now().isoformat()[:10]],
        ),
    ],
)
def test_get_days_covered_by_date_range(
    storage_locations, start_date, end_date, expected_output
):
    if start_date != datetime.min:
        start_date = parse_datetime_bound(start_date)
    if end_date != datetime.max:
        end_date = parse_datetime_bound(end_date)
    assert (
        report_data._get_days_covered_by_date_range(1, start_date, end_date)
        == expected_output
    )


@pytest.mark.parametrize(
    "cumulative",
    [
        # Differential totals
        (False),
        # Cumulative totals
        (True),
    ],
)
def test_storage_locations_usage_over_time(storage_locations, cumulative):
    """Test response from report_data.storage_locations_usage_over_time endpoint."""
    report = report_data.storage_locations_usage_over_time(
        storage_service_id=1,
        start_date=parse_datetime_bound(DATE_BEFORE_AIP_1),
        end_date=parse_datetime_bound(DATE_AFTER_AIP_3, upper=True),
        cumulative=cumulative,
    )

    assert report[fields.FIELD_STORAGE_NAME] == "test storage service"

    locations = report[fields.FIELD_LOCATIONS_USAGE_OVER_TIME]

    # Assert that every day between Jan 1, 2020 and June 1, 2021, inclusive, is
    # accounted for.
    assert len(locations) == 884

    first_day = locations["2020-01-01"]

    # Test that we have data for both storage locations.
    assert len(first_day) == 2

    # Test correctness of that data.
    first_location = first_day[0]
    assert first_location[fields.FIELD_AIPS] == 1
    assert first_location[fields.FIELD_SIZE] == 600
    assert first_location[fields.FIELD_FILE_COUNT] == 2

    second_location = first_day[1]
    assert second_location[fields.FIELD_AIPS] == 0
    assert second_location[fields.FIELD_SIZE] == 0
    assert second_location[fields.FIELD_FILE_COUNT] == 0

    last_day = locations["2021-06-01"]

    # Test that we have data for both storage locations.
    assert len(last_day) == 2

    if cumulative:
        # All AIPs in fixture should be accounted for.
        first_location = last_day[0]
        assert first_location[fields.FIELD_AIPS] == 2
        assert first_location[fields.FIELD_SIZE] == 1600
        assert first_location[fields.FIELD_FILE_COUNT] == 3

        second_location = last_day[1]
        assert second_location[fields.FIELD_AIPS] == 1
        assert second_location[fields.FIELD_SIZE] == 5000
        assert second_location[fields.FIELD_FILE_COUNT] == 2
    else:
        # Data should reflect no ingests this day.
        first_location = last_day[0]
        assert first_location[fields.FIELD_AIPS] == 0
        assert first_location[fields.FIELD_SIZE] == 0
        assert first_location[fields.FIELD_FILE_COUNT] == 0

        second_location = last_day[1]
        assert second_location[fields.FIELD_AIPS] == 0
        assert second_location[fields.FIELD_SIZE] == 0
        assert second_location[fields.FIELD_FILE_COUNT] == 0



"""AIPscan API entry point

Provides metadata to the swagger endpoint as well as importing the
different namespace modules required to provide all the functionality
we'd like from AIPscan.

HOWTO extend: Proposal is to evolve the API. So avoid versioning, and
instead, add fields to endpoints. Add endpoints. And when absolutely
necessary, following good communication with stakeholders, sunset, and
deprecate endpoints.
"""

from flask import Blueprint
from flask_restx import Api

from .namespace_data import api as ns1
from .namespace_infos import api as ns3
from .namespace_report_data import api as ns2

api_doc = """
The AIPscan API is separated into different endpoints, providing access
to the data retrieved by the application from the various Storage
Services associated with it.
"""

api = Blueprint("api", __name__, url_prefix="/api")
report_api = Api(
    api,
    title="AIPscan API endpoint - you know, for data!",
    version="0.1",
    description=api_doc,
    license="Apache License 2.0",
    license_url="https://github.com/artefactual-labs/AIPscan/blob/60afe37bfd0106735b47635606a0658a90ba9946/LICENSE",
    default_mediatype="application/json",
)

report_api.add_namespace(ns1)
report_api.add_namespace(ns2)
report_api.add_namespace(ns3)



"""Report on original copies that have a preservation derivative and
the different file formats associated with both.
"""

from flask import render_template
from flask import request

from AIPscan.Data import fields
from AIPscan.Data import report_data
from AIPscan.helpers import parse_bool
from AIPscan.Reporter import download_csv
from AIPscan.Reporter import reporter
from AIPscan.Reporter import request_params
from AIPscan.Reporter import translate_headers

HEADERS = [
    fields.FIELD_AIP_NAME,
    fields.FIELD_ORIGINAL_FILE,
    fields.FIELD_ORIGINAL_FORMAT,
    fields.FIELD_DERIVATIVE_FILE,
    fields.FIELD_DERIVATIVE_FORMAT,
]

CSV_HEADERS = [
    fields.FIELD_AIP_UUID,
    fields.FIELD_AIP_NAME,
    fields.FIELD_UUID,
    fields.FIELD_NAME,
    fields.FIELD_FORMAT,
    fields.FIELD_ORIGINAL_UUID,
    fields.FIELD_ORIGINAL_NAME,
    fields.FIELD_ORIGINAL_FORMAT,
    fields.FIELD_ORIGINAL_VERSION,
    fields.FIELD_ORIGINAL_PUID,
]


def _get_unique_aips(derivative_files):
    """Return list of AIPs represented in data.

    :param derivative files: Derivative files from report endpoint data (list
        of dicts)

    :returns: Unique AIPs with derivatives (list of dicts with keys
        fields.FIELD_AIP_UUID and fields.FIELD_AIP_NAME)
    """
    captured_keys = [fields.FIELD_AIP_UUID, fields.FIELD_AIP_NAME]
    filtered = {
        tuple((k, file_[k]) for k in sorted(file_) if k in captured_keys): file_
        for file_ in derivative_files
    }
    filtered_files = list(filtered.values())
    return [
        {
            fields.FIELD_AIP_UUID: file_[fields.FIELD_AIP_UUID],
            fields.FIELD_AIP_NAME: file_[fields.FIELD_AIP_NAME],
        }
        for file_ in filtered_files
    ]


@reporter.route("preservation_derivatives/", methods=["GET"])
def preservation_derivatives():
    """Return a report of derivative files mapped to AIPs and originals."""
    storage_service_id = request.args.get(request_params.STORAGE_SERVICE_ID)
    storage_location_id = request.args.get(request_params.STORAGE_LOCATION_ID)
    csv = parse_bool(request.args.get(request_params.CSV), default=False)
    aip_uuid = request.args.get(request_params.AIP_UUID)

    headers = translate_headers(HEADERS)

    derivative_data = report_data.preservation_derivatives(
        storage_service_id, storage_location_id, aip_uuid
    )
    derivative_files = derivative_data[fields.FIELD_FILES]

    total_files = len(derivative_files)

    unique_aips = _get_unique_aips(derivative_files)

    if csv:
        filename = "preservation_derivatives.csv"
        headers = translate_headers(CSV_HEADERS)
        return download_csv(headers, derivative_files, filename)

    return render_template(
        "report_preservation_derivatives.html",
        storage_service_id=storage_service_id,
        storage_service_name=derivative_data.get(fields.FIELD_STORAGE_NAME),
        storage_location_id=storage_location_id,
        storage_location_description=derivative_data.get(fields.FIELD_STORAGE_LOCATION),
        columns=headers,
        files=derivative_files,
        total_files=total_files,
        aips=unique_aips,
        selected_aip=aip_uuid,
    )



import datetime
import math

import typesense
from flask import current_app
from sqlalchemy import inspect

from AIPscan.models import AIP
from AIPscan.models import File
from AIPscan.models import FileType

APP_MODELS = [AIP, File]

FACET_FIELDS = {
    "file": ["file_format", "file_type", "puid", "aip_id", "size", "aip_create_date"]
}

AIP_FIELDS_TO_CACHE = {
    "storage_service_id": None,
    "storage_location_id": None,
    "create_date": "aip_create_date",
    "transfer_name": None,
    "uuid": "aip_uuid",
}


def typesense_enabled():
    return current_app.config["TYPESENSE_API_KEY"] is not None


def client():
    config = current_app.config
    return typesense.Client(
        {
            "api_key": config["TYPESENSE_API_KEY"],
            "nodes": [
                {
                    "host": config["TYPESENSE_HOST"],
                    "port": config["TYPESENSE_PORT"],
                    "protocol": config["TYPESENSE_PROTOCOL"],
                }
            ],
            "connection_timeout_seconds": int(config["TYPESENSE_TIMEOUT_SECONDS"]),
        }
    )


def search(collection, search_parameters, ts_client=None):
    if ts_client is None:
        ts_client = client()

    return ts_client.collections[collection_prefix(collection)].documents.search(
        search_parameters
    )


def collection_prefix(collection):
    return current_app.config["TYPESENSE_COLLECTION_PREFIX"] + collection


EPOCH_UTC = datetime.datetime(1970, 1, 1, tzinfo=datetime.timezone.utc)


def datetime_to_timestamp_int(datetime_obj):
    """Convert a date or datetime to a Unix timestamp (seconds).

    This function normalizes to midnight and computes seconds since the Unix
    epoch in UTC. Naive datetimes are treated as UTC. `time.mktime` was used in
    the past but avoided now because its behavior is inconsistent across
    platforms and timezones.
    """
    # Floor to midnight.
    base = datetime_obj.replace(hour=0, minute=0, second=0, microsecond=0)

    # Normalize to UTC.
    if base.tzinfo is None:
        # Naive datetime: assume it's in UTC by attaching UTC tzinfo.
        base = base.replace(tzinfo=datetime.timezone.utc)
    else:
        # Aware datetime: convert the timestamp to the equivalent UTC time.
        base = base.astimezone(datetime.timezone.utc)

    return int((base - EPOCH_UTC).total_seconds())


def assemble_filter_by(filters):
    filter_by = ""

    for clause in filters:
        if filter_by != "":
            filter_by += " && "

        filter_by += f"{clause[0]}:{clause[1]}{str(clause[2])}"

    return filter_by


def file_filters(storage_service_id, storage_location_id, start_date, end_date):
    start_timestamp = datetime_to_timestamp_int(start_date)
    end_timestamp = datetime_to_timestamp_int(end_date)

    filters = [
        ("aip_create_date", ">=", start_timestamp),
        ("aip_create_date", "<", end_timestamp),
        ("storage_service_id", "=", storage_service_id),
        ("file_type", "=", "'original'"),
    ]

    if storage_location_id is not None and storage_location_id != "":
        filters.append(("storage_location_id", "=", storage_location_id))

    return filters


def facet_value_counts(result, field_name=None):
    facet_value_counts = {}

    for facet_count in result["facet_counts"]:
        facet_value_counts[facet_count["field_name"]] = {}

        for count in facet_count["counts"]:
            facet_value_counts[facet_count["field_name"]][count["value"]] = count[
                "count"
            ]

    if field_name:
        return facet_value_counts[field_name]

    return facet_value_counts


def get_model_table(model):
    inst = inspect(model)
    return str(inst.tables[0])


def collection_fields_from_model(model):
    # Reference model definition
    inst = inspect(model)
    table = str(inst.tables[0])

    # Define Typesense fields for model
    fields = []

    for c in inst.columns:
        field = {"name": c.name}

        # Use Python type as basis of Typesense type
        try:
            py_type = c.type.python_type
        except NotImplementedError:
            py_type = str

        if py_type == FileType:
            field["type"] = "string"
        else:
            ts_types = {
                int: "int32",
                str: "string",
                bool: "bool",
                datetime.datetime: "int64",
            }

            # Use larger number type for size data
            if c.name == "size":
                ts_types[int] = "int64"

            if py_type not in ts_types:
                # Fallback to string for any remaining unmapped types
                field["type"] = "string"
            else:
                field["type"] = ts_types[py_type]

        # Mark certain fields as facets
        if table in FACET_FIELDS and field["name"] in FACET_FIELDS[table]:
            field["facet"] = True

        # Mark non-id fields as optional (to accommodate None values)
        if field["name"] != "id":
            field["optional"] = True

        fields.append(field)

    # AIP collection documents will be augmented with related data
    if table == "aip":
        fields.append({"name": "original_file_count", "type": "int32"})

    # File collection documents will be augmented with related data
    if table == "file":
        fields.append({"name": "storage_service_id", "type": "int32"})
        fields.append({"name": "storage_location_id", "type": "int32"})
        fields.append({"name": "aip_create_date", "type": "int64", "facet": True})
        fields.append({"name": "transfer_name", "type": "string"})
        fields.append({"name": "aip_uuid", "type": "string"})

    return fields


def initialize_index():
    ts_client = client()

    # Create Typesense collections containing data for each model
    for model in APP_MODELS:
        table = get_model_table(model)

        # Delete collection if it exists
        try:
            ts_client.collections[collection_prefix(table)].delete()
        except typesense.exceptions.ObjectNotFound:
            pass

        # Create collection
        fields = collection_fields_from_model(model)

        ts_client.collections.create(
            {"name": collection_prefix(table), "fields": fields}
        )


def model_instance_to_document(model, instance, model_fields=None):
    if model_fields is None:
        model_fields = collection_fields_from_model(model)

    document = {}

    for field in model_fields:
        if hasattr(instance, field["name"]):
            value = getattr(instance, field["name"])

            # Typesense IDs need to be strings
            if field["name"] == "id":
                value = str(value)

            if isinstance(value, datetime.datetime):
                value = int(value.timestamp())

            if isinstance(value, FileType):
                value = str(value).replace("FileType.", "")

            if value is not None:
                document[field["name"]] = value

    return document


def augment_file_document_with_aip_data(table, document, aip_cache, result):
    for aip_field in AIP_FIELDS_TO_CACHE:
        if aip_field != "create_date":
            if AIP_FIELDS_TO_CACHE[aip_field] is None:
                document[aip_field] = aip_cache[aip_field][result.aip_id]
            else:
                document_field = AIP_FIELDS_TO_CACHE[aip_field]
                document[document_field] = aip_cache[aip_field][result.aip_id]

    # Round AIP create date to the day
    dt = aip_cache["create_date"][result.aip_id]
    dt = datetime.datetime(*dt.timetuple()[0:3])
    document["aip_create_date"] = int(dt.strftime("%s"))


def populate_index():
    ts_client = client()

    # Initialize AIP cache
    aip_cache = {}
    for aip_field in AIP_FIELDS_TO_CACHE:
        aip_cache[aip_field] = {}

    # Add documents to collections
    for model in APP_MODELS:
        table = get_model_table(model)
        model_fields = collection_fields_from_model(model)

        results = model.query.all()

        docs = []
        indexed = 0
        completed_percent = 0

        for result in results:
            document = model_instance_to_document(model, result, model_fields)

            if model is File:
                augment_file_document_with_aip_data(table, document, aip_cache, result)

            docs.append(document)
            perform_bulk_document_creation(ts_client, table, docs)

            indexed += 1
            completed_percent = math.ceil(indexed / len(results) * 100)

            status = {
                "type": model.__name__,
                "indexed": indexed,
                "total": len(results),
                "percent": completed_percent,
            }

            if model == AIP:
                for aip_field in AIP_FIELDS_TO_CACHE:
                    aip_cache[aip_field][result.id] = getattr(result, aip_field)

            yield status

        finish_bulk_document_creation(ts_client, table, docs)


def perform_bulk_document_creation(ts_client, table, docs):
    # Send bulk creation every 40 documents
    if len(docs) == 40:
        ts_client.collections[collection_prefix(table)].documents.import_(docs)
        docs = []


def finish_bulk_document_creation(ts_client, table, docs):
    # Send bulk creation of remaining documents
    if len(docs):
        ts_client.collections[collection_prefix(table)].documents.import_(docs)



import pytest
from flask import current_app
from werkzeug.datastructures import Headers

from AIPscan import test_helpers
from AIPscan.Reporter import views


def test_download_mets(app_with_populated_files, mocker):
    # Mock storage server API request response
    request = mocker.patch("requests.get")

    class MockResponse:
        pass

    return_response = MockResponse()
    return_response.content = bytes("METS content", "utf-8")
    return_response.status_code = 200

    request.return_value = return_response

    # Test view logic
    with current_app.test_client() as test_client:
        response = test_client.get("/reporter/download_mets/1")

        assert response.headers == Headers(
            [
                (
                    "Content-Disposition",
                    'attachment; filename="METS-111111111111-1111-1111-11111111.xml"',
                ),
                ("Content-Length", "12"),
                ("Content-Type", "text/html; charset=utf-8"),
            ]
        )

        assert response.data == return_response.content
        assert response.status_code == return_response.status_code


@pytest.mark.parametrize(
    "query,page,pages,total,prev_num,next_num,set_location",
    [
        (None, 1, 2, 3, None, 2, False),  # No search query
        ("Another Test AIP", 1, 1, 1, None, None, False),  # Search matches one AIP
        (None, 1, 1, 1, None, None, True),  # Use non-default storage location
        ("Does Not Exist", 1, 0, 0, None, None, False),  # Search doesn't match any AIP
        ("Test AIP", 1, 2, 3, None, 2, False),  # Search matches all AIPs, page 1
        ("Test AIP", 2, 2, 3, 1, None, False),  # Search matches all AIPs, page 2
    ],
)
def test_get_aip_pager(
    app_instance,
    mocker,
    query,
    page,
    pages,
    total,
    prev_num,
    next_num,
    set_location,
):
    storage_service = test_helpers.create_test_storage_service()
    default_storage_location = test_helpers.create_test_storage_location(
        default_storage_service_id=storage_service.id
    )
    other_storage_location = test_helpers.create_test_storage_location(
        storage_service_id=storage_service.id,
        current_location="/other/location",
    )

    test_helpers.create_test_aip(
        transfer_name="Test AIP",
        storage_service_id=storage_service.id,
        storage_location_id=default_storage_location.id,
    )
    test_helpers.create_test_aip(
        transfer_name="Another Test AIP",
        storage_service_id=storage_service.id,
        storage_location_id=default_storage_location.id,
    )
    test_helpers.create_test_aip(
        transfer_name="This Is Also A Test AIP",
        storage_service_id=storage_service.id,
        storage_location_id=other_storage_location.id,
    )

    location = None
    if set_location:
        location = other_storage_location

    pager = views.get_aip_pager(
        page, 2, storage_service, storage_location=location, query=query
    )

    assert pager.pages == pages
    assert pager.total == total

    if prev_num is None:
        assert pager.prev_num is None
    else:
        assert pager.prev_num == prev_num

    if next_num is None:
        assert pager.next_num is None
    else:
        assert pager.next_num == next_num


@pytest.mark.parametrize("page,pager_page", [(1, 1), ("bad", 1)])
def test_get_file_pager(app_instance, mocker, page, pager_page):
    paginate_mock = mocker.Mock()
    filter_by_mock = mocker.Mock()
    filter_by_mock.paginate.return_value = paginate_mock

    query_mock = mocker.Mock()
    query_mock.filter_by.return_value = filter_by_mock

    aip = test_helpers.create_test_aip()
    pager = views.get_file_pager(page, 2, aip)

    assert pager.page == pager_page
    assert pager.pages == 0
    assert pager.per_page == 2
    assert pager.prev_num is None
    assert pager.next_num is None


def test_view_aip(app_with_populated_files):
    with current_app.test_client() as test_client:
        response = test_client.get("/reporter/aip/1")
        assert response.status_code == 200

        response = test_client.get("/reporter/aip/0")
        assert response.status_code == 404


def test_view_file(app_with_populated_files):
    with current_app.test_client() as test_client:
        response = test_client.get("/reporter/file/1")
        assert response.status_code == 200

        response = test_client.get("/reporter/file/0")
        assert response.status_code == 404



from flask import current_app

EXPECTED_CSV_CONTENTS = b"UUID,Filename,Size,Size (bytes),Type,Format,Version,PUID,AIP Name,AIP UUID\r\n555555555555-5555-5555-55555555,preservation.jpg,2.0 kB,2000,original,JPEG,1.02,fmt/44,Test AIP,222222222222-2222-2222-22222222\r\n333333333333-3333-3333-33333333,original.jpg,1.0 kB,1000,original,JPEG,1.01,fmt/43,Test AIP,111111111111-1111-1111-11111111\r\n444444444444-4444-4444-44444444,original.iso,0 Bytes,0,original,ISO Disk Image File,,fmt/468,Test AIP,222222222222-2222-2222-22222222\r\n"


def test_largest_files(app_with_populated_format_versions):
    """Test that report template renders."""
    with current_app.test_client() as test_client:
        response = test_client.get("/reporter/largest_files/?amss_id=1")
        assert response.status_code == 200


def test_largest_files_csv(app_with_populated_format_versions):
    """Test CSV export."""
    with current_app.test_client() as test_client:
        response = test_client.get("/reporter/largest_files/?amss_id=1&csv=True")
        assert response.status_code == 200
        assert (
            response.headers["Content-Disposition"]
            == "attachment; filename=largest_files.csv"
        )
        assert response.mimetype == "text/csv"
        assert response.data == EXPECTED_CSV_CONTENTS
